#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from copy import deepcopy

from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType
from libnmstate.schema import InterfaceState
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceIPv6

from libnmstate.ifaces.ifaces import Ifaces

from .constants import FOO_IFACE_NAME
from .constants import IPV4_ADDRESSES
from .constants import IPV6_ADDRESSES


def gen_foo_iface_info(iface_type=InterfaceType.ETHERNET):
    return {
        Interface.NAME: FOO_IFACE_NAME,
        Interface.TYPE: iface_type,
        Interface.STATE: InterfaceState.UP,
        Interface.IPV4: {InterfaceIPv4.ENABLED: False},
        Interface.IPV6: {InterfaceIPv6.ENABLED: False},
    }


def gen_foo_iface_info_static_ip(iface_type=InterfaceType.ETHERNET):
    iface_info = gen_foo_iface_info(iface_type)
    iface_info.update(
        {
            Interface.IPV4: {
                InterfaceIPv4.ENABLED: True,
                InterfaceIPv4.DHCP: False,
                InterfaceIPv4.ADDRESS: deepcopy(IPV4_ADDRESSES),
            },
            Interface.IPV6: {
                InterfaceIPv6.ENABLED: True,
                InterfaceIPv6.DHCP: False,
                InterfaceIPv6.ADDRESS: deepcopy(IPV6_ADDRESSES),
                InterfaceIPv6.AUTOCONF: False,
            },
        }
    )
    return iface_info


def gen_two_static_ip_ifaces(iface1_name, iface2_name):
    iface1_info = gen_foo_iface_info_static_ip()
    iface1_info[Interface.NAME] = iface1_name
    iface1_info[Interface.IPV4][InterfaceIPv4.ADDRESS].pop(1)
    iface1_info[Interface.IPV6][InterfaceIPv6.ADDRESS].pop(1)
    iface2_info = gen_foo_iface_info_static_ip()
    iface2_info[Interface.NAME] = iface2_name
    iface2_info[Interface.IPV4][InterfaceIPv4.ADDRESS].pop(0)
    iface2_info[Interface.IPV6][InterfaceIPv6.ADDRESS].pop(0)
    return Ifaces([], [iface1_info, iface2_info])



#
# Copyright (c) 2019-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import os

import pytest

import libnmstate
from libnmstate.schema import Ethernet
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceState

from .testlib import assertlib
from .testlib import statelib


MAC1 = "00:11:22:33:44:55"
MAC2 = "00:11:22:33:44:66"
MAC3 = "00:11:22:33:44:FF"
MAC_MIX_CASE = "00:11:22:33:44:Ff"

VF0_CONF = {
    Ethernet.SRIOV.VFS.ID: 0,
    Ethernet.SRIOV.VFS.SPOOF_CHECK: True,
    Ethernet.SRIOV.VFS.MAC_ADDRESS: MAC1,
    Ethernet.SRIOV.VFS.TRUST: False,
}

VF1_CONF = {
    Ethernet.SRIOV.VFS.ID: 1,
    Ethernet.SRIOV.VFS.SPOOF_CHECK: True,
    Ethernet.SRIOV.VFS.MAC_ADDRESS: MAC2,
    Ethernet.SRIOV.VFS.TRUST: False,
}


def _test_nic_name():
    return os.environ.get("TEST_REAL_NIC")


@pytest.fixture
def disable_sriov():
    pf_name = _test_nic_name()
    iface_info = {
        Interface.NAME: pf_name,
        Interface.STATE: InterfaceState.UP,
        Ethernet.CONFIG_SUBTREE: {
            Ethernet.SRIOV_SUBTREE: {
                Ethernet.SRIOV.TOTAL_VFS: 0,
                Ethernet.SRIOV.VFS_SUBTREE: [],
            }
        },
    }
    desired_state = {Interface.KEY: [iface_info]}
    libnmstate.apply(desired_state)
    yield
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: pf_name,
                    Interface.STATE: InterfaceState.ABSENT,
                }
            ]
        }
    )


@pytest.fixture
def sriov_interface(disable_sriov):
    pf_name = _test_nic_name()
    iface_info = {
        Interface.NAME: pf_name,
        Interface.STATE: InterfaceState.UP,
        Ethernet.CONFIG_SUBTREE: {
            Ethernet.SRIOV_SUBTREE: {Ethernet.SRIOV.TOTAL_VFS: 2},
        },
    }
    desired_state = {Interface.KEY: [iface_info]}
    libnmstate.apply(desired_state)
    yield desired_state


@pytest.fixture
def sriov_iface_vf(disable_sriov):
    pf_name = _test_nic_name()
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: pf_name,
                Ethernet.CONFIG_SUBTREE: {
                    Ethernet.SRIOV_SUBTREE: {
                        Ethernet.SRIOV.TOTAL_VFS: 2,
                        Ethernet.SRIOV.VFS_SUBTREE: [VF0_CONF, VF1_CONF],
                    }
                },
            }
        ]
    }
    libnmstate.apply(desired_state)
    yield desired_state


@pytest.mark.skipif(
    not os.environ.get("TEST_REAL_NIC"),
    reason="Need to define TEST_REAL_NIC for SR-IOV test",
)
def test_sriov_with_no_vfs_config(sriov_interface):
    assertlib.assert_state_match(sriov_interface)


@pytest.mark.skipif(
    not os.environ.get("TEST_REAL_NIC"),
    reason="Need to define TEST_REAL_NIC for SR-IOV test",
)
def test_sriov_increase_vfs(sriov_interface):
    desired_state = sriov_interface
    eth_config = desired_state[Interface.KEY][0][Ethernet.CONFIG_SUBTREE]
    eth_config[Ethernet.SRIOV_SUBTREE][Ethernet.SRIOV.TOTAL_VFS] = 5
    libnmstate.apply(desired_state)
    assertlib.assert_state_match(desired_state)


@pytest.mark.skipif(
    not os.environ.get("TEST_REAL_NIC"),
    reason="Need to define TEST_REAL_NIC for SR-IOV test",
)
def test_sriov_decrease_vfs(sriov_interface):
    desired_state = sriov_interface
    eth_config = desired_state[Interface.KEY][0][Ethernet.CONFIG_SUBTREE]
    eth_config[Ethernet.SRIOV_SUBTREE][Ethernet.SRIOV.TOTAL_VFS] = 1
    eth_config[Ethernet.SRIOV_SUBTREE][Ethernet.SRIOV.VFS_SUBTREE] = [VF0_CONF]
    libnmstate.apply(desired_state)
    assertlib.assert_state_match(desired_state)


@pytest.mark.skipif(
    not os.environ.get("TEST_REAL_NIC"),
    reason="Need to define TEST_REAL_NIC for SR-IOV test",
)
def test_sriov_create_vf_config(sriov_iface_vf):
    assertlib.assert_state_match(sriov_iface_vf)


@pytest.mark.skipif(
    not os.environ.get("TEST_REAL_NIC"),
    reason="Need to define TEST_REAL_NIC for SR-IOV test",
)
def test_sriov_edit_vf_config(sriov_iface_vf):
    desired_state = sriov_iface_vf
    eth_config = desired_state[Interface.KEY][0][Ethernet.CONFIG_SUBTREE]
    vf0 = eth_config[Ethernet.SRIOV_SUBTREE][Ethernet.SRIOV.VFS_SUBTREE][0]
    vf0[Ethernet.SRIOV.VFS.TRUST] = True
    vf0[Ethernet.SRIOV.VFS.MAC_ADDRESS] = MAC3
    libnmstate.apply(desired_state)
    assertlib.assert_state_match(desired_state)


@pytest.mark.skipif(
    not os.environ.get("TEST_REAL_NIC"),
    reason="Need to define TEST_REAL_NIC for SR-IOV test",
)
@pytest.mark.xfail(
    raises=libnmstate.error.NmstateVerificationError,
    reason="https://github.com/nmstate/nmstate/issues/1454",
    strict=True,
)
def test_sriov_remove_vf_config(sriov_iface_vf):
    desired_state = sriov_iface_vf
    eth_config = desired_state[Interface.KEY][0][Ethernet.CONFIG_SUBTREE]
    eth_config[Ethernet.SRIOV_SUBTREE][Ethernet.SRIOV.VFS_SUBTREE] = []
    libnmstate.apply(desired_state)
    assertlib.assert_state_match(desired_state)


@pytest.mark.skipif(
    not os.environ.get("TEST_REAL_NIC"),
    reason="Need to define TEST_REAL_NIC for SR-IOV test",
)
def test_sriov_vf_mac_mixed_case(sriov_iface_vf):
    desired_state = sriov_iface_vf
    eth_config = desired_state[Interface.KEY][0][Ethernet.CONFIG_SUBTREE]
    vf0 = eth_config[Ethernet.SRIOV_SUBTREE][Ethernet.SRIOV.VFS_SUBTREE][0]
    vf0[Ethernet.SRIOV.VFS.MAC_ADDRESS] = MAC_MIX_CASE
    libnmstate.apply(desired_state)

    vf0[Ethernet.SRIOV.VFS.MAC_ADDRESS] = MAC_MIX_CASE.upper()
    assertlib.assert_state_match(desired_state)


@pytest.mark.skipif(
    not os.environ.get("TEST_REAL_NIC"),
    reason="Need to define TEST_REAL_NIC for SR-IOV test",
)
def test_wait_sriov_vf_been_created():
    pf_name = _test_nic_name()
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: pf_name,
                Ethernet.CONFIG_SUBTREE: {
                    Ethernet.SRIOV_SUBTREE: {Ethernet.SRIOV.TOTAL_VFS: 2}
                },
            }
        ]
    }
    try:
        libnmstate.apply(desired_state)
        assertlib.assert_state_match(desired_state)
        current_state = statelib.show_only((f"{pf_name}v0", f"{pf_name}v1"))
        assert len(current_state[Interface.KEY]) == 2

    finally:
        desired_state[Interface.KEY][0][
            Interface.STATE
        ] = InterfaceState.ABSENT
        libnmstate.apply(desired_state)


@pytest.mark.skipif(
    not os.environ.get("TEST_REAL_NIC"),
    reason="Need to define TEST_REAL_NIC for SR-IOV test",
)
def test_show_saved_config_for_sriov_down(sriov_iface_vf):
    pf_name = _test_nic_name()
    running_state = statelib.show_only((pf_name,))
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: pf_name,
                    Interface.STATE: InterfaceState.DOWN,
                }
            ]
        }
    )
    saved_state = statelib.show_saved_config_only((pf_name,))
    iface_state = saved_state[Interface.KEY][0]

    assert saved_state[Interface.KEY][0][Interface.STATE] == InterfaceState.UP
    assert iface_state[Ethernet.SRIOV_SUBTREE][Ethernet.SRIOV.TOTAL_VFS] == 2
    assertlib.assert_state_match_full(saved_state, running_state)



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import pytest
from unittest import mock

from libnmstate import nm
from libnmstate.error import NmstateValueError
from libnmstate.ifaces.bond import BondIface
from libnmstate.schema import Bond
from libnmstate.schema import BondMode
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType


@pytest.fixture
def nm_mock():
    with mock.patch.object(nm.bond, "NM") as m:
        yield m


def _gen_bond_iface(bond_options, mode=BondMode.ROUND_ROBIN):
    return BondIface(
        {
            Interface.NAME: "foo",
            Interface.TYPE: InterfaceType.BOND,
            Bond.CONFIG_SUBTREE: {
                Bond.MODE: mode,
                Bond.OPTIONS_SUBTREE: bond_options,
            },
        }
    )


def test_create_setting(nm_mock):
    bond_setting_mock = nm_mock.SettingBond.new.return_value
    bond_setting_mock.add_option.return_value = True

    iface = _gen_bond_iface({"miimon": "100"})
    nm.bond.create_setting(iface, wired_setting=None, base_con_profile=None)

    bond_setting_mock.add_option.assert_has_calls(
        [
            mock.call(Bond.MODE, BondMode.ROUND_ROBIN),
            mock.call("miimon", "100"),
        ],
        any_order=True,
    )


def test_create_setting_with_invalid_bond_option(nm_mock):
    bond_setting_mock = nm_mock.SettingBond.new.return_value
    bond_setting_mock.add_option.return_value = False

    iface = _gen_bond_iface({"foo": "100"})

    with pytest.raises(NmstateValueError):
        nm.bond.create_setting(
            iface, wired_setting=None, base_con_profile=None
        )


def test_create_setting_with_mac_restriction(nm_mock):
    bond_setting_mock = nm_mock.SettingBond.new.return_value
    bond_setting_mock.add_option.return_value = True

    iface = _gen_bond_iface(
        {"fail_over_mac": "active"}, BondMode.ACTIVE_BACKUP
    )

    wired_setting = mock.MagicMock()
    wired_setting.props.cloned_mac_address = "02:ff:ff:ff:ff:01"

    nm.bond.create_setting(iface, wired_setting, base_con_profile=None)
    assert wired_setting.props.cloned_mac_address is None



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
import io
import json
import subprocess

from unittest import mock

from nmstatectl import nmstatectl

LO_JSON_STATE = """{
    "route-rules": {
        "config": []
    },
    "routes": {
        "config": [],
        "running": []
    },
    "interfaces": [
        {
            "name": "lo",
            "type": "unknown",
            "state": "down"
        }
    ]
}
"""

EMPTY_JSON_STATE = """{
    "route-rules": {
        "config": []
    },
    "routes": {
        "config": [],
        "running": []
    },
    "interfaces": []
}
"""

EMPTY_YAML_STATE = """---
route-rules:
  config: []
routes:
  config: []
  running: []
interfaces: []
"""

LO_YAML_STATE = """---
route-rules:
  config: []
routes:
  config: []
  running: []
interfaces:
- name: lo
  type: unknown
  state: down
"""


def _mock_libnmstate_apply_func(
    state,
    verify_change=True,
    commit=True,
    rollback_timeout=60,
    save_to_disk=True,
):
    return None


@mock.patch("sys.argv", ["nmstatectl", "set", "mystate.json"])
@mock.patch.object(nmstatectl.libnmstate, "apply", _mock_libnmstate_apply_func)
@mock.patch.object(
    nmstatectl, "open", mock.mock_open(read_data="{}"), create=True
)
def test_run_ctl_directly_set():
    nmstatectl.main()


@mock.patch("sys.argv", ["nmstatectl", "show"])
@mock.patch.object(nmstatectl.libnmstate, "show", lambda: {})
def test_run_ctl_directly_show_empty():
    nmstatectl.main()


@mock.patch("sys.argv", ["nmstatectl", "show", "non_existing_interface"])
@mock.patch.object(
    nmstatectl.libnmstate, "show", lambda: json.loads(LO_JSON_STATE)
)
@mock.patch("nmstatectl.nmstatectl.sys.stdout", new_callable=io.StringIO)
def test_run_ctl_directly_show_only_empty(mock_stdout):
    nmstatectl.main()
    assert mock_stdout.getvalue() == EMPTY_YAML_STATE


@mock.patch("sys.argv", ["nmstatectl", "show", "lo"])
@mock.patch.object(
    nmstatectl.libnmstate, "show", lambda: json.loads(LO_JSON_STATE)
)
@mock.patch("nmstatectl.nmstatectl.sys.stdout", new_callable=io.StringIO)
def test_run_ctl_directly_show_only(mock_stdout):
    nmstatectl.main()
    assert mock_stdout.getvalue() == LO_YAML_STATE


@mock.patch(
    "sys.argv", ["nmstatectl", "show", "--json", "non_existing_interface"]
)
@mock.patch.object(
    nmstatectl.libnmstate, "show", lambda: json.loads(LO_JSON_STATE)
)
@mock.patch("nmstatectl.nmstatectl.sys.stdout", new_callable=io.StringIO)
def test_run_ctl_directly_show_json_only_empty(mock_stdout):
    nmstatectl.main()
    assert mock_stdout.getvalue() == EMPTY_JSON_STATE


@mock.patch("sys.argv", ["nmstatectl", "show", "--json", "lo"])
@mock.patch.object(
    nmstatectl.libnmstate, "show", lambda: json.loads(LO_JSON_STATE)
)
@mock.patch("nmstatectl.nmstatectl.sys.stdout", new_callable=io.StringIO)
def test_run_ctl_directly_show_json_only(mock_stdout):
    nmstatectl.main()
    assert mock_stdout.getvalue() == LO_JSON_STATE


def test_run_ctl_executable():
    rc = subprocess.call(["nmstatectl", "--help"])
    assert rc == 0



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
from copy import deepcopy

from libnmstate.state import merge_dict


class TestMergeDict:
    def test_merge_empty(self):
        dict_to = {}
        dict_from = {}
        merge_dict(dict_to, dict_from)

        assert dict_to == {}
        assert dict_from == {}

    def test_merge_simple_dict(self):
        dict_to = {"a": 1}
        dict_from = {"a": 0, "b": "b"}
        merge_dict(dict_to, dict_from)

        assert dict_to == {"a": 1, "b": "b"}

    def test_merge_dict_with_list_value(self):
        dict_to = {"a": [1, 2, 3], "b": []}
        dict_from = {"a": [1, 3], "b": [3], "c": 5}

        merge_dict(dict_to, dict_from)

        assert dict_to == {"a": [1, 2, 3], "b": [], "c": 5}

    def test_merge_nested_dict(self):
        dict_to = {
            "a": {
                "a.1": {
                    "a.1.1": 1,
                    "a.1.2": "2",
                    "a.1.3": [3, 4, 5, 6, 7, 8, 9],
                }
            }
        }
        dict_from = {
            "a": {
                "a.1": {"a.1.3": [3], "a.1.4": [4, 5], "a.1.5": {}},
                "a.2": [1, 2, 3],
            },
            "b": 2,
        }

        merge_dict(dict_to, dict_from)

        assert dict_to == {
            "a": {
                "a.1": {
                    "a.1.1": 1,
                    "a.1.2": "2",
                    "a.1.3": [3, 4, 5, 6, 7, 8, 9],
                    "a.1.4": [4, 5],
                    "a.1.5": {},
                },
                "a.2": [1, 2, 3],
            },
            "b": 2,
        }

    def test_merge_with_from_unchanged(self):
        dict_to = {"a": 1, "b": "1"}
        dict_from = {"a": 2, "b": "2", "c": 3}
        expected_from = deepcopy(dict_from)

        merge_dict(dict_to, dict_from)

        assert dict_from == expected_from
        assert dict_to == {"a": 1, "b": "1", "c": 3}

    def test_merge_with_to_dict_contains_none_as_value(self):
        dict_to = {"a": 1, "b": None}
        dict_from = {"a": 2, "b": "2", "c": 3}
        expected_from = deepcopy(dict_from)

        merge_dict(dict_to, dict_from)

        assert dict_from == expected_from
        assert dict_to == {"a": 1, "b": None, "c": 3}

    def test_merge_with_from_dict_contains_none_as_value(self):
        dict_to = {"a": 1, "b": 3}
        dict_from = {"a": 2, "b": None, "c": None}
        expected_from = deepcopy(dict_from)

        merge_dict(dict_to, dict_from)

        assert dict_from == expected_from
        assert dict_to == {"a": 1, "b": 3, "c": None}



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from collections.abc import Mapping
from collections.abc import Sequence
import contextlib
import copy
from ipaddress import ip_address
from operator import itemgetter

import libnmstate
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIP
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceIPv6
from libnmstate.schema import InterfaceState
from libnmstate.schema import Bond
from libnmstate.schema import LinuxBridge
from libnmstate.schema import OVSBridge


def show_only(ifnames):
    """
    Report the current state, filtering based on the given interface names.
    """
    base_filter_state = {
        Interface.KEY: [{Interface.NAME: ifname} for ifname in ifnames]
    }
    current_state = State(libnmstate.show())
    current_state.filter(base_filter_state)
    return current_state.state


def show_saved_config_only(ifnames):
    """
    Report the saved config, filtering based on the given interface names.
    """
    base_filter_state = {
        Interface.KEY: [{Interface.NAME: ifname} for ifname in ifnames]
    }
    current_state = State(libnmstate.show_saved_config())
    current_state.filter(base_filter_state)
    return current_state.state


class State:
    def __init__(self, state):
        self._state = copy.deepcopy(state)

    def __eq__(self, other):
        return self._state == other

    def __hash__(self):
        return hash(self._state)

    def __str__(self):
        return self._state

    def __repr__(self):
        return self.__str__()

    @property
    def state(self):
        return self._state

    def filter(self, based_on_state):
        """
        Given the based_on_state, update the sate with a filtered version
        which includes only entities such as interfaces that have been
        mentioned in the based_on_state.
        In case there are no entities for filtering, all are reported.
        """
        base_iface_names = {
            ifstate[Interface.NAME]
            for ifstate in based_on_state[Interface.KEY]
        }

        if not base_iface_names:
            return

        filtered_iface_state = [
            ifstate
            for ifstate in self._state[Interface.KEY]
            if ifstate[Interface.NAME] in base_iface_names
        ]
        self._state = {Interface.KEY: filtered_iface_state}

    def update(self, other_state):
        """
        Given the other_state, update the state with the other_state data.
        """
        other_state = copy.deepcopy(other_state)
        other_interfaces_state = other_state[Interface.KEY]

        for base_iface_state in self._state[Interface.KEY]:
            ifname = base_iface_state[Interface.NAME]
            other_iface_state = _lookup_iface_state_by_name(
                other_interfaces_state, ifname
            )
            if other_iface_state is not None:
                iface_state = _dict_update(base_iface_state, other_iface_state)
                other_iface_state.update(iface_state)

        self._state = other_state

    def normalize(self):
        self._convert_lag_numeric_value_options_to_integer()
        self._sort_iface_lag_port()
        self._sort_iface_bridge_ports()
        self._ipv4_skeleton_canonicalization()
        self._ipv6_skeleton_canonicalization()
        self._ignore_dhcp_manual_addr()
        self._ignore_dhcp_option_when_off()
        self._ignore_ipv6_link_local()
        self._sort_ip_addresses()
        self._sort_interfaces_by_name()
        self._canonicalize_iface_ipv6_addresses()
        self._normalize_linux_bridge_port_vlan()
        self._upper_linux_bridge_group_addr()
        self._sort_ovs_lag_ports()

    def match(self, other):
        return _state_match(self.state, other.state)

    def _sort_interfaces_by_name(self):
        self._state[Interface.KEY].sort(key=lambda d: d[Interface.NAME])

    def _canonicalize_iface_ipv6_addresses(self):
        for iface_state in self.state[Interface.KEY]:
            iface_ipv6_state = iface_state.get(Interface.IPV6)
            if iface_ipv6_state:
                iface_ipv6_addresses_state = iface_ipv6_state.get(
                    InterfaceIP.ADDRESS
                )
                if iface_ipv6_addresses_state:
                    iface_ipv6_state[InterfaceIP.ADDRESS] = [
                        _canonicalize_ipv6_addr(iface_ipv6_addr)
                        for iface_ipv6_addr in iface_ipv6_addresses_state
                    ]

    def remove_absent_entries(self):
        self._state[Interface.KEY] = [
            ifstate
            for ifstate in self._state[Interface.KEY]
            if ifstate.get(Interface.STATE) != InterfaceState.ABSENT
        ]

    def _sort_iface_lag_port(self):
        for ifstate in self._state[Interface.KEY]:
            ifstate.get(Bond.CONFIG_SUBTREE, {}).get(Bond.PORT, []).sort()

    def _sort_iface_bridge_ports(self):
        for ifstate in self._state[Interface.KEY]:
            ifstate.get(LinuxBridge.CONFIG_SUBTREE, {}).get(
                LinuxBridge.PORT_SUBTREE, []
            ).sort(key=itemgetter(LinuxBridge.Port.NAME))

    def _ipv6_skeleton_canonicalization(self):
        for iface_state in self._state.get(Interface.KEY, []):
            iface_state.setdefault(Interface.IPV6, {})
            iface_state[Interface.IPV6].setdefault(
                InterfaceIPv6.ENABLED, False
            )
            iface_state[Interface.IPV6].setdefault(InterfaceIPv6.ADDRESS, [])
            iface_state[Interface.IPV6].setdefault(InterfaceIPv6.DHCP, False)
            iface_state[Interface.IPV6].setdefault(
                InterfaceIPv6.AUTOCONF, False
            )

    def _ipv4_skeleton_canonicalization(self):
        for iface_state in self._state.get(Interface.KEY, []):
            iface_state.setdefault(Interface.IPV4, {})
            iface_state[Interface.IPV4].setdefault(
                InterfaceIPv4.ENABLED, False
            )
            iface_state[Interface.IPV4].setdefault(InterfaceIPv4.ADDRESS, [])
            iface_state[Interface.IPV4].setdefault(InterfaceIPv4.DHCP, False)

    def _ignore_ipv6_link_local(self):
        for iface_state in self._state.get(Interface.KEY, []):
            iface_state[Interface.IPV6][InterfaceIPv6.ADDRESS] = list(
                addr
                for addr in iface_state[Interface.IPV6][InterfaceIPv6.ADDRESS]
                if not _is_ipv6_link_local(
                    addr[InterfaceIPv6.ADDRESS_IP],
                    addr[InterfaceIPv6.ADDRESS_PREFIX_LENGTH],
                )
            )

    def _sort_ip_addresses(self):
        for iface_state in self._state.get(Interface.KEY, []):
            for family in (Interface.IPV4, Interface.IPV6):
                iface_state.get(family, {}).get(InterfaceIP.ADDRESS, []).sort(
                    key=itemgetter(InterfaceIP.ADDRESS_IP)
                )

    def _ignore_dhcp_manual_addr(self):
        for iface_state in self._state.get(Interface.KEY, []):
            for family in (Interface.IPV4, Interface.IPV6):
                if iface_state.get(family, {}).get(InterfaceIP.DHCP):
                    iface_state[family][InterfaceIP.ADDRESS] = []

    def _ignore_dhcp_option_when_off(self):
        for iface_state in self._state.get(Interface.KEY, []):
            for family in (Interface.IPV4, Interface.IPV6):
                ip = iface_state.get(family, {})
                if not (
                    ip.get(InterfaceIP.ENABLED)
                    and (
                        ip.get(InterfaceIP.DHCP)
                        or ip.get(InterfaceIPv6.AUTOCONF)
                    )
                ):
                    for dhcp_option in (
                        InterfaceIP.AUTO_ROUTES,
                        InterfaceIP.AUTO_GATEWAY,
                        InterfaceIP.AUTO_DNS,
                    ):
                        ip.pop(dhcp_option, None)

    def _convert_lag_numeric_value_options_to_integer(self):
        for iface_state in self._state.get(Interface.KEY, []):
            if iface_state.get(Interface.TYPE) == Bond.KEY:
                bond_state = iface_state.get(Bond.CONFIG_SUBTREE, {})
                options = bond_state.get(Bond.OPTIONS_SUBTREE)
                if options:
                    for option_name, option_value in options.items():
                        with contextlib.suppress(ValueError):
                            option_value = int(option_value)
                        options[option_name] = option_value

    def _normalize_linux_bridge_port_vlan(self):
        linux_bridges = (
            iface
            for iface in self._state.get(Interface.KEY, [])
            if iface[Interface.TYPE] == LinuxBridge.TYPE
        )
        for lb in linux_bridges:
            ports = lb.get(LinuxBridge.CONFIG_SUBTREE, {}).get(
                LinuxBridge.PORT_SUBTREE, []
            )
            for port in ports:
                if not port.get(LinuxBridge.Port.VLAN_SUBTREE):
                    port[LinuxBridge.Port.VLAN_SUBTREE] = {}

    def _upper_linux_bridge_group_addr(self):
        linux_bridges = (
            iface
            for iface in self._state.get(Interface.KEY, [])
            if iface[Interface.TYPE] == LinuxBridge.TYPE
        )
        for lb in linux_bridges:
            cur_group_addr = (
                lb.get(LinuxBridge.CONFIG_SUBTREE, {})
                .get(LinuxBridge.OPTIONS_SUBTREE, {})
                .get(LinuxBridge.Options.GROUP_ADDR)
            )
            if cur_group_addr:
                lb[LinuxBridge.CONFIG_SUBTREE][LinuxBridge.OPTIONS_SUBTREE][
                    LinuxBridge.Options.GROUP_ADDR
                ] = cur_group_addr.upper()

    def _sort_ovs_lag_ports(self):
        for iface_state in self._state[Interface.KEY]:
            for port_config in iface_state.get(
                OVSBridge.CONFIG_SUBTREE, {}
            ).get(OVSBridge.PORT_SUBTREE, []):
                port_config.get(
                    OVSBridge.Port.LINK_AGGREGATION_SUBTREE, {}
                ).get(OVSBridge.Port.LinkAggregation.PORT_SUBTREE, []).sort(
                    key=itemgetter(OVSBridge.Port.LinkAggregation.Port.NAME)
                )


def _lookup_iface_state_by_name(interfaces_state, ifname):
    for iface_state in interfaces_state:
        if iface_state[Interface.NAME] == ifname:
            return iface_state
    return None


def _dict_update(origin_data, to_merge_data):
    """
    Recursively performs a dict update (merge), taking the to_merge_data and
    updating the origin_data.
    The function changes the origin_data in-place.
    """
    if not to_merge_data:
        return to_merge_data

    for key, val in to_merge_data.items():
        if isinstance(val, Mapping):
            origin_data[key] = _dict_update(origin_data.get(key, {}), val)
        else:
            origin_data[key] = val
    return origin_data


def filter_current_state(desired_state):
    """
    Given the desired state, return a filtered version of the current state
    which includes only entities such as interfaces that have been mentioned
    in the desired state.
    In case there are no entities for filtering, all are reported.
    """
    current_state = libnmstate.show()
    desired_iface_names = {
        ifstate[Interface.NAME] for ifstate in desired_state[Interface.KEY]
    }

    if not desired_iface_names:
        return current_state

    filtered_iface_current_state = [
        ifstate
        for ifstate in current_state[Interface.KEY]
        if ifstate[Interface.NAME] in desired_iface_names
    ]
    return {Interface.KEY: filtered_iface_current_state}


def _is_ipv6_link_local(ip, prefix):
    """
    The IPv6 link local address range is fe80::/10.
    """
    return ip[:3] in ["fe8", "fe9", "fea", "feb"] and prefix >= 10


def _state_match(desire, current):
    if isinstance(desire, Mapping):
        return isinstance(current, Mapping) and all(
            _state_match(val, current.get(key)) for key, val in desire.items()
        )
    elif isinstance(desire, Sequence) and not isinstance(desire, str):
        return (
            isinstance(current, Sequence)
            and not isinstance(current, str)
            and len(current) == len(desire)
            and all(_state_match(d, c) for d, c in zip(desire, current))
        )
    else:
        return desire == current


def _canonicalize_ipv6_addr(addr):
    address = addr[InterfaceIP.ADDRESS_IP]
    if _is_ipv6_address(address):
        addr[InterfaceIP.ADDRESS_IP] = str(ip_address(address))
    return addr


def _is_ipv6_address(addr):
    return ":" in addr



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import time


def retry_till_true_or_timeout(timeout, func, *args, **kwargs):
    ret = func(*args, **kwargs)
    while timeout > 0:
        if ret:
            break
        time.sleep(1)
        timeout -= 1
        ret = func(*args, **kwargs)
    return ret



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from contextlib import contextmanager
import errno
import logging
import os

import libnmstate
import libnmstate.error as libnmError

try:
    import varlink
except ModuleNotFoundError:
    raise libnmError.NmstateDependencyError("python3 varlink module not found")


class NmstateVarlinkLogHandler(logging.Handler):
    def __init__(self):
        self._log_records = list()
        super().__init__()

    def filter(self, record):
        return True

    def emit(self, record):
        self._log_records.append(record)

    @property
    def logs(self):
        """
        Return a list of dict, example:
            [
                {
                    "time": "2003-07-08 16:49:45,896",
                    "level": "DEBUG",
                    "message": "foo is changed",
                }
            ]
        """
        return [
            {
                "time": record.asctime,
                "level": record.levelname,
                "message": record.message,
            }
            for record in self._log_records
        ]


@contextmanager
def nmstate_varlink_logger():
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    handler = NmstateVarlinkLogHandler()
    handler.setLevel(logging.DEBUG)
    logger.addHandler(handler)
    try:
        yield handler
    finally:
        logger.removeHandler(handler)


def validate_method_arguments(user_args, method_args):
    """
    Returns dictionary with validated arguments values.
    """
    kwargs = {}
    for arg in user_args.keys():
        if arg not in method_args or user_args[arg] == {}:
            raise varlink.InvalidParameter(arg)
        if user_args[arg] is not None:
            kwargs[arg] = user_args[arg]
    return kwargs


def gen_varlink_server(address):
    """
    Returns varlink server object
    Checks if the varlink address already in use.
    """
    address = "unix:" + address
    server = varlink.ThreadingServer(
        address, ServiceRequestHandler, bind_and_activate=False
    )
    server.allow_reuse_address = True
    try:
        server.server_bind()
        server.server_activate()
    except OSError as exception:
        server.shutdown()
        server.server_close()
        if exception.errno == errno.EADDRINUSE:
            server = varlink.ThreadingServer(
                address, ServiceRequestHandler, bind_and_activate=True
            )
        else:
            logging.error(exception.strerror)
            raise exception
    return server


def start_varlink_server(address):
    """
    Runs the varlink server in the specified the file path
    """
    varlink_server = gen_varlink_server(address)
    try:
        with varlink_server as server:
            server.serve_forever()
    except Exception as exception:
        logging.error(str(exception))
        varlink_server.shutdown()
    finally:
        varlink_server.server_close()


class NmstateError(varlink.VarlinkError):
    def __init__(self, message, logs):
        varlink.VarlinkError.__init__(
            self,
            {
                "error": self.__class__.__name__,
                "parameters": {
                    "error_message": message,
                    "log": logs,
                },
            },
        )


class NmstateValueError(NmstateError):
    pass


class NmstatePermissionError(NmstateError):
    pass


class NmstateConflictError(NmstateError):
    pass


class NmstateLibnmError(NmstateError):
    pass


class NmstateVerificationError(NmstateError):
    pass


class NmstateNotImplementedError(NmstateError):
    pass


class NmstateInternalError(NmstateError):
    pass


class NmstateDependencyError(NmstateError):
    pass


class ServiceRequestHandler(varlink.RequestHandler):
    service = varlink.Service(
        vendor="Red Hat",
        product="Nmstate",
        version=libnmstate.__version__,
        url="https://www.nmstate.io",
        interface_dir=os.path.dirname(__file__),
        namespaced=False,
    )


@ServiceRequestHandler.service.interface("io.nmstate")
class NmstateVarlinkService:
    def Show(self, arguments):
        """
        Reports the state data on the system
        """
        with nmstate_varlink_logger() as log_handler:
            method_args = ["include_status_data"]
            show_kwargs = validate_method_arguments(arguments, method_args)
            try:
                configured_state = libnmstate.show(**show_kwargs)
                return {"state": configured_state, "log": log_handler.logs}
            except libnmstate.error.NmstateValueError as exception:
                logging.error(str(exception))
                raise NmstateValueError(str(exception), log_handler.logs)

    def ShowRunningConfig(self, arguments):
        with nmstate_varlink_logger() as log_handler:
            method_args = []
            validate_method_arguments(arguments, method_args)
            try:
                configured_state = libnmstate.show_running_config()
                return {"state": configured_state, "log": log_handler.logs}
            except libnmstate.error.NmstateValueError as exception:
                logging.error(str(exception))
                raise NmstateValueError(str(exception), log_handler.logs)

    def ShowSavedConfig(self, arguments):
        with nmstate_varlink_logger() as log_handler:
            method_args = []
            validate_method_arguments(arguments, method_args)
            try:
                configured_state = libnmstate.show_saved_config()
                return {"state": configured_state, "log": log_handler.logs}
            except libnmstate.error.NmstateValueError as exception:
                logging.error(str(exception))
                raise NmstateValueError(str(exception), log_handler.logs)

    def Apply(self, arguments):
        """
        Apply desired state declared in json format
        which is parsed as dictionary
        """
        with nmstate_varlink_logger() as log_handler:
            method_args = [
                "desired_state",
                "verify_change",
                "commit",
                "rollback_timeout",
                "save_to_disk",
            ]
            apply_kwargs = validate_method_arguments(arguments, method_args)
            if "desired_state" not in apply_kwargs.keys():
                logging.error("Desired_state not specified")
                raise NmstateValueError(
                    "desired_state: No state specified", log_handler.logs
                )
            try:
                libnmstate.apply(**apply_kwargs)
                return {"log": log_handler.logs}
            except TypeError as exception:
                logging.error(str(exception), log_handler.logs)
                raise varlink.InvalidParameter(exception)
            except libnmstate.error.NmstatePermissionError as exception:
                logging.error(str(exception))
                raise NmstatePermissionError(str(exception), log_handler.logs)
            except libnmstate.error.NmstateValueError as exception:
                logging.error(str(exception))
                raise NmstateValueError(str(exception), log_handler.logs)
            except libnmstate.error.NmstateConflictError as exception:
                logging.error(str(exception))
                raise NmstateConflictError(str(exception), log_handler.logs)
            except libnmstate.error.NmstateVerificationError as exception:
                logging.error(str(exception))
                raise NmstateVerificationError(
                    str(exception), log_handler.logs
                )

    def Commit(self, arguments):
        """
        Commits the checkpoint
        """
        with nmstate_varlink_logger() as log_handler:
            method_args = ["checkpoint"]
            commit_kwargs = validate_method_arguments(arguments, method_args)
            try:
                libnmstate.commit(**commit_kwargs)
                return {"log": log_handler.logs}
            except libnmstate.error.NmstateValueError as exception:
                logging.error(str(exception))
                raise NmstateValueError(str(exception), log_handler.logs)

    def Rollback(self, arguments):
        """
        Roll back to the checkpoint
        """
        with nmstate_varlink_logger() as log_handler:
            method_args = ["checkpoint"]
            rollback_kwargs = validate_method_arguments(arguments, method_args)
            try:
                libnmstate.rollback(**rollback_kwargs)
                return {"log": log_handler.logs}
            except libnmstate.error.NmstateValueError as exception:
                logging.error(str(exception))
                raise NmstateValueError(str(exception), log_handler.logs)



#
# Copyright (c) 2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from contextlib import contextmanager

import libnmstate

from libnmstate.schema import Interface
from libnmstate.schema import InterfaceState
from libnmstate.schema import VXLAN


class VxlanState:
    def __init__(self, id, base_if, remote, destination_port=4789):
        self.name = f"{base_if}.{id}"
        self.id = id
        self.base_if = base_if
        self.remote = remote
        self.destination_port = destination_port

    @property
    def up(self):
        return {
            Interface.NAME: self.name,
            Interface.TYPE: VXLAN.TYPE,
            Interface.STATE: InterfaceState.UP,
            VXLAN.CONFIG_SUBTREE: {
                VXLAN.ID: self.id,
                VXLAN.BASE_IFACE: self.base_if,
                VXLAN.REMOTE: self.remote,
                VXLAN.DESTINATION_PORT: self.destination_port,
            },
        }

    @property
    def absent(self):
        return {
            Interface.NAME: self.name,
            Interface.TYPE: VXLAN.TYPE,
            Interface.STATE: InterfaceState.ABSENT,
        }

    @property
    def down(self):
        return {
            Interface.NAME: self.name,
            Interface.TYPE: VXLAN.TYPE,
            Interface.STATE: InterfaceState.DOWN,
        }


@contextmanager
def vxlan_interfaces(*vxlans, create=True):
    setup_state = vxlans_up(vxlans)
    if create:
        libnmstate.apply(setup_state)
    try:
        yield setup_state
    finally:
        libnmstate.apply(vxlans_absent(vxlans), verify_change=False)


def vxlans_up(vxlans):
    vxlans_state = [vxlan_state.up for vxlan_state in vxlans]
    return {Interface.KEY: vxlans_state}


def vxlans_absent(vxlans):
    vxlans_state = [vxlan_state.absent for vxlan_state in vxlans]
    return {Interface.KEY: vxlans_state}


def vxlans_down(vxlans):
    vxlans_state = [vxlan_state.down for vxlan_state in vxlans]
    return {Interface.KEY: vxlans_state}



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import pytest

import libnmstate
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceIPv6
from libnmstate.schema import InterfaceState
from libnmstate.schema import LinuxBridge as LB

from ..testlib.bridgelib import linux_bridge
from ..testlib.dummy import nm_unmanaged_dummy


BRIDGE0 = "brtest0"
DUMMY1 = "dummy1"


@pytest.fixture
def nm_unmanaged_dummy1():
    with nm_unmanaged_dummy(DUMMY1):
        yield


@pytest.mark.tier1
def test_bridge_as_port_unmanaged_interface(nm_unmanaged_dummy1):
    with linux_bridge(
        BRIDGE0,
        bridge_subtree_state={
            LB.OPTIONS_SUBTREE: {LB.STP_SUBTREE: {LB.STP.ENABLED: False}}
        },
    ) as desired_state:
        bridge_iface_state = desired_state[Interface.KEY][0]
        bridge_iface_state[LB.CONFIG_SUBTREE] = {
            LB.PORT_SUBTREE: [{LB.Port.NAME: DUMMY1}],
        }
        # To reproduce bug https://bugzilla.redhat.com/1816517
        # explitly define dummy1 as IPv4/IPv6 disabled is required.
        desired_state[Interface.KEY].append(
            {
                Interface.NAME: DUMMY1,
                Interface.STATE: InterfaceState.UP,
                Interface.IPV4: {InterfaceIPv4.ENABLED: False},
                Interface.IPV6: {InterfaceIPv6.ENABLED: False},
            }
        )
        libnmstate.apply(desired_state)



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import pytest

import libnmstate
from libnmstate.iplib import is_ipv6_link_local_addr
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceIPv6
from libnmstate.schema import InterfaceState
from libnmstate.schema import InterfaceType

from .testlib import assertlib
from .testlib import cmdlib
from .testlib import statelib
from .testlib.dummy import nm_unmanaged_dummy
from .testlib.iproutelib import ip_monitor_assert_stable_link_up

# TEST-NET addresses: https://tools.ietf.org/html/rfc5737#section-3
IPV4_ADDRESS1 = "192.0.2.251"
IPV4_ADDRESS2 = "192.0.2.252"
IPV4_ADDRESS3 = "198.51.100.249"
IPV4_ADDRESS4 = "198.51.100.250"
# IPv6 Address Prefix Reserved for Documentation:
# https://tools.ietf.org/html/rfc3849
IPV6_ADDRESS1 = "2001:db8:1::1"
IPV6_ADDRESS2 = "2001:db8:2::1"
IPV6_LINK_LOCAL_ADDRESS1 = "fe80::1"
IPV6_LINK_LOCAL_ADDRESS2 = "fe80::2"

DUMMY1 = "dummy1"


@pytest.fixture
def setup_eth1_ipv4(eth1_up):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: "eth1",
                Interface.TYPE: InterfaceType.ETHERNET,
                Interface.STATE: InterfaceState.UP,
                Interface.IPV4: {
                    InterfaceIPv4.ENABLED: True,
                    InterfaceIPv4.ADDRESS: [
                        {
                            InterfaceIPv4.ADDRESS_IP: IPV4_ADDRESS1,
                            InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
                        }
                    ],
                },
            }
        ]
    }
    libnmstate.apply(desired_state)
    yield desired_state


@pytest.fixture
def setup_eth1_ipv6(eth1_up):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: "eth1",
                Interface.TYPE: InterfaceType.ETHERNET,
                Interface.STATE: InterfaceState.UP,
                Interface.IPV6: {
                    InterfaceIPv6.ENABLED: True,
                    InterfaceIPv6.ADDRESS: [
                        {
                            InterfaceIPv6.ADDRESS_IP: IPV6_ADDRESS1,
                            InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
                        }
                    ],
                },
            }
        ]
    }
    libnmstate.apply(desired_state)

    return desired_state


@pytest.fixture
def setup_eth1_ipv6_disable(eth1_up):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: "eth1",
                Interface.TYPE: InterfaceType.ETHERNET,
                Interface.STATE: InterfaceState.UP,
                Interface.IPV6: {InterfaceIPv6.ENABLED: False},
            }
        ]
    }
    libnmstate.apply(desired_state)

    return desired_state


def test_add_static_ipv4_with_full_state(eth1_up):
    desired_state = statelib.show_only(("eth1",))
    eth1_desired_state = desired_state[Interface.KEY][0]

    eth1_desired_state[Interface.STATE] = InterfaceState.UP
    eth1_desired_state[Interface.IPV4][InterfaceIPv4.ENABLED] = True
    eth1_desired_state[Interface.IPV4][InterfaceIPv4.ADDRESS] = [
        {
            InterfaceIPv4.ADDRESS_IP: IPV4_ADDRESS3,
            InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
        }
    ]
    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)


@pytest.mark.tier1
def test_add_static_ipv4_with_min_state(eth2_up):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: "eth2",
                Interface.TYPE: InterfaceType.ETHERNET,
                Interface.STATE: InterfaceState.UP,
                Interface.IPV4: {
                    InterfaceIPv4.ENABLED: True,
                    InterfaceIPv4.ADDRESS: [
                        {
                            InterfaceIPv4.ADDRESS_IP: IPV4_ADDRESS4,
                            InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
                        }
                    ],
                },
            }
        ]
    }
    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)


@pytest.mark.tier1
def test_remove_static_ipv4(setup_eth1_ipv4):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: "eth1",
                Interface.TYPE: InterfaceType.ETHERNET,
                Interface.IPV4: {InterfaceIPv4.ENABLED: False},
            }
        ]
    }

    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)


@pytest.mark.tier1
def test_edit_static_ipv4_address_and_prefix(setup_eth1_ipv4):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: "eth1",
                Interface.TYPE: InterfaceType.ETHERNET,
                Interface.STATE: InterfaceState.UP,
                Interface.IPV4: {
                    InterfaceIPv4.ENABLED: True,
                    InterfaceIPv4.ADDRESS: [
                        {
                            InterfaceIPv4.ADDRESS_IP: IPV4_ADDRESS2,
                            InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 30,
                        }
                    ],
                },
            }
        ]
    }

    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)


def test_add_ifaces_with_same_static_ipv4_address_in_one_transaction(
    eth1_up, eth2_up
):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: "eth1",
                Interface.TYPE: InterfaceType.ETHERNET,
                Interface.STATE: InterfaceState.UP,
                Interface.IPV4: {
                    InterfaceIPv4.ENABLED: True,
                    InterfaceIPv4.ADDRESS: [
                        {
                            InterfaceIPv4.ADDRESS_IP: IPV4_ADDRESS1,
                            InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
                        }
                    ],
                },
            },
            {
                Interface.NAME: "eth2",
                Interface.TYPE: InterfaceType.ETHERNET,
                Interface.STATE: InterfaceState.UP,
                Interface.IPV4: {
                    InterfaceIPv4.ENABLED: True,
                    InterfaceIPv4.ADDRESS: [
                        {
                            InterfaceIPv4.ADDRESS_IP: IPV4_ADDRESS1,
                            InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
                        }
                    ],
                },
            },
        ]
    }

    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)


@pytest.mark.tier1
def test_add_iface_with_same_static_ipv4_address_to_existing(
    setup_eth1_ipv4, eth2_up
):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: "eth2",
                Interface.TYPE: InterfaceType.ETHERNET,
                Interface.STATE: InterfaceState.UP,
                Interface.IPV4: {
                    InterfaceIPv4.ENABLED: True,
                    InterfaceIPv4.ADDRESS: [
                        {
                            InterfaceIPv4.ADDRESS_IP: IPV4_ADDRESS1,
                            InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
                        }
                    ],
                },
            }
        ]
    }
    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)


@pytest.mark.tier1
def test_add_static_ipv6_with_full_state(eth1_up):
    desired_state = statelib.show_only(("eth1",))
    eth1_desired_state = desired_state[Interface.KEY][0]
    eth1_desired_state[Interface.STATE] = InterfaceState.UP
    eth1_desired_state[Interface.IPV6][InterfaceIPv6.ENABLED] = True
    eth1_desired_state[Interface.IPV6][InterfaceIPv6.ADDRESS] = [
        {
            InterfaceIPv6.ADDRESS_IP: IPV6_ADDRESS2,
            InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
        },
        # This sequence is intentionally made for IP address sorting.
        {
            InterfaceIPv6.ADDRESS_IP: IPV6_ADDRESS1,
            InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
        },
    ]
    libnmstate.apply(desired_state)
    assertlib.assert_state(desired_state)


def test_add_static_ipv6_with_link_local(eth1_up):
    desired_state = statelib.show_only(("eth1",))
    eth1_desired_state = desired_state[Interface.KEY][0]
    eth1_desired_state[Interface.STATE] = InterfaceState.UP
    eth1_desired_state[Interface.IPV6][InterfaceIPv6.ENABLED] = True
    eth1_desired_state[Interface.IPV6][InterfaceIPv6.ADDRESS] = [
        {
            InterfaceIPv6.ADDRESS_IP: IPV6_LINK_LOCAL_ADDRESS1,
            InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
        },
        {
            InterfaceIPv6.ADDRESS_IP: IPV6_ADDRESS1,
            InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
        },
    ]

    libnmstate.apply(desired_state)

    # Make sure only the link local address got ignored.
    cur_state = statelib.show_only(("eth1",))
    eth1_cur_state = cur_state[Interface.KEY][0]
    assert (
        eth1_desired_state[Interface.IPV6][InterfaceIPv6.ADDRESS][0]
        not in eth1_cur_state[Interface.IPV6][InterfaceIPv6.ADDRESS]
    )
    assert (
        eth1_desired_state[Interface.IPV6][InterfaceIPv6.ADDRESS][1]
        in eth1_cur_state[Interface.IPV6][InterfaceIPv6.ADDRESS]
    )


def test_add_static_ipv6_with_link_local_only(eth1_up):
    desired_state = statelib.show_only(("eth1",))
    eth1_desired_state = desired_state[Interface.KEY][0]
    eth1_desired_state[Interface.STATE] = InterfaceState.UP
    eth1_desired_state[Interface.IPV6][InterfaceIPv6.ENABLED] = True
    eth1_desired_state[Interface.IPV6][InterfaceIPv6.ADDRESS] = [
        {
            InterfaceIPv6.ADDRESS_IP: IPV6_LINK_LOCAL_ADDRESS1,
            InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
        },
        {
            InterfaceIPv6.ADDRESS_IP: IPV6_LINK_LOCAL_ADDRESS2,
            InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
        },
    ]

    libnmstate.apply(desired_state)

    # Make sure the link local address got ignored.
    cur_state = statelib.show_only(("eth1",))
    eth1_cur_state = cur_state[Interface.KEY][0]
    assert (
        eth1_desired_state[Interface.IPV6][InterfaceIPv6.ADDRESS][0]
        not in eth1_cur_state[Interface.IPV6][InterfaceIPv6.ADDRESS]
    )
    assert (
        eth1_desired_state[Interface.IPV6][InterfaceIPv6.ADDRESS][1]
        not in eth1_cur_state[Interface.IPV6][InterfaceIPv6.ADDRESS]
    )


@pytest.mark.tier1
def test_add_static_ipv6_with_no_address(eth1_up):
    desired_state = statelib.show_only(("eth1",))
    eth1_desired_state = desired_state[Interface.KEY][0]
    eth1_desired_state[Interface.STATE] = InterfaceState.UP
    eth1_desired_state[Interface.IPV6][InterfaceIPv6.ENABLED] = True

    libnmstate.apply(desired_state)

    cur_state = statelib.show_only(("eth1",))
    eth1_cur_state = cur_state[Interface.KEY][0]
    # Should have at least 1 link-local address.
    assert len(eth1_cur_state[Interface.IPV6][InterfaceIPv6.ADDRESS]) >= 1


def test_add_static_ipv6_with_min_state(eth2_up):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: "eth2",
                Interface.TYPE: InterfaceType.ETHERNET,
                Interface.STATE: InterfaceState.UP,
                Interface.IPV6: {
                    InterfaceIPv6.ENABLED: True,
                    InterfaceIPv6.ADDRESS: [
                        {
                            InterfaceIPv6.ADDRESS_IP: IPV6_ADDRESS1,
                            InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
                        }
                    ],
                },
            }
        ]
    }
    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)


@pytest.mark.tier1
def test_disable_static_ipv6(setup_eth1_ipv6):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: "eth1",
                Interface.TYPE: InterfaceType.ETHERNET,
                Interface.IPV6: {InterfaceIPv6.ENABLED: False},
            }
        ]
    }

    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)


@pytest.mark.tier1
def test_disable_static_ipv6_and_rollback(setup_eth1_ipv6):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: "eth1",
                Interface.TYPE: InterfaceType.ETHERNET,
                Interface.IPV6: {InterfaceIPv6.ENABLED: False},
                "foo": "bad_value",
            }
        ]
    }

    with pytest.raises(libnmstate.error.NmstateVerificationError):
        libnmstate.apply(desired_state)

    assertlib.assert_state(setup_eth1_ipv6)


@pytest.mark.tier1
def test_enable_ipv6_and_rollback_to_disable_ipv6(setup_eth1_ipv6_disable):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: "eth1",
                Interface.TYPE: InterfaceType.ETHERNET,
                Interface.IPV6: {
                    InterfaceIPv6.ENABLED: True,
                    InterfaceIPv6.ADDRESS: [
                        {
                            InterfaceIPv6.ADDRESS_IP: IPV6_ADDRESS1,
                            InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
                        }
                    ],
                },
                "foo": "bad_value",
            }
        ]
    }

    with pytest.raises(libnmstate.error.NmstateVerificationError):
        libnmstate.apply(desired_state)

    assertlib.assert_state(setup_eth1_ipv6_disable)


@pytest.mark.tier1
def test_edit_static_ipv6_address_and_prefix(setup_eth1_ipv6):
    eth1_setup = setup_eth1_ipv6[Interface.KEY][0]
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: "eth1",
                Interface.TYPE: InterfaceType.ETHERNET,
                Interface.STATE: InterfaceState.UP,
                Interface.IPV6: {
                    InterfaceIPv6.ENABLED: True,
                    InterfaceIPv6.ADDRESS: [
                        {
                            InterfaceIPv6.ADDRESS_IP: IPV6_ADDRESS2,
                            InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
                        }
                    ],
                },
            }
        ]
    }

    libnmstate.apply(desired_state)
    eth1_desired_state = desired_state[Interface.KEY][0]
    current_state = statelib.show_only(("eth1",))

    eth1_current_state = current_state[Interface.KEY][0]

    assert (
        eth1_desired_state[Interface.IPV6][InterfaceIPv6.ADDRESS][0]
        in eth1_current_state[Interface.IPV6][InterfaceIPv6.ADDRESS]
    )

    assert (
        eth1_setup[Interface.IPV6][InterfaceIPv6.ADDRESS][0]
        not in eth1_current_state[Interface.IPV6][InterfaceIPv6.ADDRESS]
    )


def test_add_ifaces_with_same_static_ipv6_address_in_one_transaction(
    eth1_up, eth2_up
):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: "eth1",
                Interface.TYPE: InterfaceType.ETHERNET,
                Interface.STATE: InterfaceState.UP,
                Interface.IPV6: {
                    InterfaceIPv6.ENABLED: True,
                    InterfaceIPv6.ADDRESS: [
                        {
                            InterfaceIPv6.ADDRESS_IP: IPV6_ADDRESS1,
                            InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
                        }
                    ],
                },
            },
            {
                Interface.NAME: "eth2",
                Interface.TYPE: InterfaceType.ETHERNET,
                Interface.STATE: InterfaceState.UP,
                Interface.IPV6: {
                    InterfaceIPv6.ENABLED: True,
                    InterfaceIPv6.ADDRESS: [
                        {
                            InterfaceIPv6.ADDRESS_IP: IPV6_ADDRESS1,
                            InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
                        }
                    ],
                },
            },
        ]
    }

    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)


def test_add_iface_with_same_static_ipv6_address_to_existing(
    setup_eth1_ipv6, eth2_up
):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: "eth2",
                Interface.TYPE: InterfaceType.ETHERNET,
                Interface.STATE: InterfaceState.UP,
                Interface.IPV6: {
                    InterfaceIPv6.ENABLED: True,
                    InterfaceIPv6.ADDRESS: [
                        {
                            InterfaceIPv6.ADDRESS_IP: IPV6_ADDRESS1,
                            InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
                        }
                    ],
                },
            }
        ]
    }
    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)


def test_add_iface_with_static_ipv6_expanded_format(eth1_up):
    ipv6_addr_lead_zeroes = "2001:0db8:85a3:0000:0000:8a2e:0370:7331"
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: eth1_up[Interface.KEY][0][Interface.NAME],
                Interface.TYPE: InterfaceType.ETHERNET,
                Interface.STATE: InterfaceState.UP,
                Interface.IPV6: {
                    InterfaceIPv6.ENABLED: True,
                    InterfaceIPv6.ADDRESS: [
                        {
                            InterfaceIPv6.ADDRESS_IP: ipv6_addr_lead_zeroes,
                            InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
                        }
                    ],
                },
            }
        ]
    }
    libnmstate.apply(desired_state)
    assertlib.assert_state(desired_state)


@pytest.mark.tier1
@ip_monitor_assert_stable_link_up("eth1")
def test_modify_ipv4_with_reapply(setup_eth1_ipv4):
    ipv4_addr = IPV4_ADDRESS2
    ipv4_state = setup_eth1_ipv4[Interface.KEY][0][Interface.IPV4]
    ipv4_state[InterfaceIPv4.ADDRESS][0][InterfaceIPv4.ADDRESS_IP] = ipv4_addr
    libnmstate.apply(setup_eth1_ipv4)

    assertlib.assert_state(setup_eth1_ipv4)


@pytest.mark.tier1
@ip_monitor_assert_stable_link_up("eth1")
def test_modify_ipv6_with_reapply(setup_eth1_ipv6):
    ipv6_addr = IPV6_ADDRESS2
    ipv6_state = setup_eth1_ipv6[Interface.KEY][0][Interface.IPV6]
    ipv6_state[InterfaceIPv6.ADDRESS][0][InterfaceIPv6.ADDRESS_IP] = ipv6_addr
    libnmstate.apply(setup_eth1_ipv6)

    assertlib.assert_state(setup_eth1_ipv6)


@pytest.mark.tier1
def test_get_ip_address_from_unmanaged_dummy():
    with nm_unmanaged_dummy(DUMMY1):
        cmdlib.exec_cmd(
            f"ip addr add {IPV4_ADDRESS1}/24 dev {DUMMY1}".split(), check=True
        )
        cmdlib.exec_cmd(
            f"ip -6 addr add {IPV6_ADDRESS2}/64 dev {DUMMY1}".split(),
            check=True,
        )
        iface_state = statelib.show_only((DUMMY1,))[Interface.KEY][0]
        # Remove IPv6 link local address
        iface_state[Interface.IPV6][InterfaceIPv6.ADDRESS] = [
            addr
            for addr in iface_state[Interface.IPV6][InterfaceIPv6.ADDRESS]
            if not is_ipv6_link_local_addr(
                addr[InterfaceIPv6.ADDRESS_IP],
                addr[InterfaceIPv6.ADDRESS_PREFIX_LENGTH],
            )
        ]

        assert iface_state[Interface.IPV4] == {
            InterfaceIPv4.ENABLED: True,
            InterfaceIPv4.ADDRESS: [
                {
                    InterfaceIPv4.ADDRESS_IP: IPV4_ADDRESS1,
                    InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
                }
            ],
        }

        assert iface_state[Interface.IPV6] == {
            InterfaceIPv6.ENABLED: True,
            InterfaceIPv6.ADDRESS: [
                {
                    InterfaceIPv6.ADDRESS_IP: IPV6_ADDRESS2,
                    InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
                }
            ],
        }


def test_ignore_invalid_ip_on_absent_interface(eth1_up):
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: "eth1",
                    Interface.STATE: InterfaceState.ABSENT,
                    Interface.IPV4: {
                        InterfaceIPv4.ENABLED: True,
                        InterfaceIPv4.ADDRESS: [
                            {
                                InterfaceIPv4.ADDRESS_IP: "a.a.a.a",
                                InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
                            }
                        ],
                    },
                    Interface.IPV6: {
                        InterfaceIPv6.ENABLED: True,
                        InterfaceIPv6.ADDRESS: [
                            {
                                InterfaceIPv6.ADDRESS_IP: "::g",
                                InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
                            }
                        ],
                    },
                }
            ]
        }
    )



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from collections.abc import Mapping
from collections.abc import Sequence
from copy import deepcopy
import difflib
import json

import yaml

from .schema import DNS
from .schema import Route
from .schema import RouteRule
from .schema import Interface

PRIORITY_LIST = (
    "name",
    "type",
    "state",
    "enabled",
    DNS.KEY,
    RouteRule.KEY,
    Route.KEY,
    Interface.KEY,
)


def format_desired_current_state_diff(desired_state, current_state):
    pretty_desired_state = PrettyState(desired_state).yaml
    pretty_current_state = PrettyState(current_state).yaml

    diff = "".join(
        difflib.unified_diff(
            pretty_desired_state.splitlines(True),
            pretty_current_state.splitlines(True),
            fromfile="desired",
            tofile="current",
            n=3,
        )
    )
    return (
        "\n"
        "desired\n"
        "=======\n"
        "{}\n"
        "current\n"
        "=======\n"
        "{}\n"
        "difference\n"
        "==========\n"
        "{}\n".format(pretty_desired_state, pretty_current_state, diff)
    )


class PrettyState:
    def __init__(self, state):
        yaml.add_representer(dict, represent_dict)
        self.state = _sort_with_priority(state)

    @property
    def yaml(self):
        return yaml.dump(
            self.state, default_flow_style=False, explicit_start=True
        )

    @property
    def json(self):
        return json.dumps(self.state, indent=4, separators=(",", ": "))


def represent_dict(dumper, data):
    """
    Represent dictionary with insert order
    """
    value = []

    for item_key, item_value in data.items():
        node_key = dumper.represent_data(item_key)
        node_value = dumper.represent_data(item_value)
        value.append((node_key, node_value))

    return yaml.nodes.MappingNode("tag:yaml.org,2002:map", value)


def represent_unicode(_, data):
    """
    Represent unicode as regular string

    Source:
        https://stackoverflow.com/questions/1950306/pyyaml-dumping-without-tags

    """

    return yaml.ScalarNode(
        tag="tag:yaml.org,2002:str", value=data.encode("utf-8")
    )


def _sort_with_priority(data):
    if isinstance(data, Sequence) and not isinstance(data, str):
        return [_sort_with_priority(item) for item in data]
    elif isinstance(data, Mapping):
        new_data = {}
        for key in sorted(data.keys(), key=_sort_with_priority_key_func):
            new_data[key] = _sort_with_priority(data[key])
        return new_data
    else:
        return deepcopy(data)


def _sort_with_priority_key_func(key):
    try:
        priority = PRIORITY_LIST.index(key)
    except ValueError:
        priority = len(PRIORITY_LIST)
    return (priority, key)



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import logging
import socket

from libnmstate import iplib
from libnmstate.error import NmstateNotImplementedError
from libnmstate.nm import dns as nm_dns
from libnmstate.nm import route as nm_route
from libnmstate.schema import InterfaceIPv6
from libnmstate.schema import Route

from ..ifaces import BaseIface
from .common import NM

IPV6_DEFAULT_ROUTE_METRIC = 1024
INT32_MAX = 2 ** 31 - 1


def get_ipv6_config(nm_profile, full_config=False):
    """
    If full_config is False, only provides information regarding:
        * Enable status
        * DHCP/Autoconf status
    """
    if nm_profile is None:
        # Neither unmanaged or not active, let nispor determine its state
        return {}

    info = {
        InterfaceIPv6.ENABLED: False,
        InterfaceIPv6.DHCP: False,
        InterfaceIPv6.AUTOCONF: False,
    }

    ip_profile = nm_profile.get_setting_ip6_config() if nm_profile else None
    if ip_profile:
        info[InterfaceIPv6.ENABLED] = True
        method = ip_profile.get_method()
        if method == NM.SETTING_IP6_CONFIG_METHOD_AUTO:
            info[InterfaceIPv6.DHCP] = True
            info[InterfaceIPv6.AUTOCONF] = True
        elif method == NM.SETTING_IP6_CONFIG_METHOD_DHCP:
            info[InterfaceIPv6.DHCP] = True
            info[InterfaceIPv6.AUTOCONF] = False
        elif method == NM.SETTING_IP6_CONFIG_METHOD_DISABLED:
            info[InterfaceIPv6.ENABLED] = False

        if info[InterfaceIPv6.DHCP] or info[InterfaceIPv6.AUTOCONF]:
            props = ip_profile.props
            info[InterfaceIPv6.AUTO_ROUTES] = not props.ignore_auto_routes
            info[InterfaceIPv6.AUTO_GATEWAY] = not props.never_default
            info[InterfaceIPv6.AUTO_DNS] = not props.ignore_auto_dns
            info[InterfaceIPv6.AUTO_ROUTE_TABLE_ID] = props.route_table
        if full_config:
            info[InterfaceIPv6.ADDRESS] = []
            for nm_addr in ip_profile.props.addresses:
                info[InterfaceIPv6.ADDRESS].append(
                    {
                        InterfaceIPv6.ADDRESS_IP: nm_addr.get_address(),
                        InterfaceIPv6.ADDRESS_PREFIX_LENGTH: int(
                            nm_addr.get_prefix()
                        ),
                    }
                )

    return info


def create_setting(config, base_con_profile):
    setting_ip = None
    if base_con_profile and config and config.get(InterfaceIPv6.ENABLED):
        setting_ip = base_con_profile.get_setting_ip6_config()
        if setting_ip:
            setting_ip = setting_ip.duplicate()
            setting_ip.clear_addresses()
            setting_ip.props.ignore_auto_routes = False
            setting_ip.props.never_default = False
            setting_ip.props.ignore_auto_dns = False
            setting_ip.clear_routes()
            setting_ip.clear_routing_rules()
            setting_ip.props.gateway = None
            setting_ip.props.route_table = Route.USE_DEFAULT_ROUTE_TABLE
            setting_ip.props.route_metric = Route.USE_DEFAULT_METRIC
            setting_ip.clear_dns()
            setting_ip.clear_dns_searches()
            setting_ip.props.dns_priority = nm_dns.DEFAULT_DNS_PRIORITY

    if not setting_ip:
        setting_ip = NM.SettingIP6Config.new()

    # Ensure IPv6 RA and DHCPv6 is based on MAC address only
    setting_ip.props.addr_gen_mode = NM.SettingIP6ConfigAddrGenMode.EUI64
    setting_ip.props.dhcp_duid = "ll"
    setting_ip.props.dhcp_iaid = "mac"

    if not config or not config.get(InterfaceIPv6.ENABLED):
        setting_ip.props.method = NM.SETTING_IP6_CONFIG_METHOD_DISABLED
        return setting_ip

    is_dhcp = config.get(InterfaceIPv6.DHCP, False)
    is_autoconf = config.get(InterfaceIPv6.AUTOCONF, False)
    ip_addresses = config.get(InterfaceIPv6.ADDRESS, ())

    if is_dhcp or is_autoconf:
        _set_dynamic(setting_ip, is_dhcp, is_autoconf)
        # NetworkManager will remove the virtual interface when DHCPv6 or
        # IPv6-RA timeout, set them to infinity.
        setting_ip.props.dhcp_timeout = INT32_MAX
        setting_ip.props.ra_timeout = INT32_MAX
        setting_ip.props.ignore_auto_routes = not config.get(
            InterfaceIPv6.AUTO_ROUTES, True
        )
        setting_ip.props.never_default = not config.get(
            InterfaceIPv6.AUTO_GATEWAY, True
        )
        setting_ip.props.ignore_auto_dns = not config.get(
            InterfaceIPv6.AUTO_DNS, True
        )
        route_table = config.get(InterfaceIPv6.AUTO_ROUTE_TABLE_ID)
        if route_table:
            setting_ip.props.route_table = route_table

    elif ip_addresses:
        _set_static(setting_ip, ip_addresses)
    else:
        setting_ip.props.method = NM.SETTING_IP6_CONFIG_METHOD_LINK_LOCAL

    nm_route.add_routes(setting_ip, config.get(BaseIface.ROUTES_METADATA, []))
    nm_dns.add_dns(setting_ip, config.get(BaseIface.DNS_METADATA, {}))
    nm_route.add_route_rules(
        setting_ip,
        socket.AF_INET6,
        config.get(BaseIface.ROUTE_RULES_METADATA, []),
    )
    return setting_ip


def _set_dynamic(setting_ip, is_dhcp, is_autoconf):
    if not is_dhcp and is_autoconf:
        raise NmstateNotImplementedError(
            "Autoconf without DHCP is not supported yet"
        )

    if is_dhcp and is_autoconf:
        setting_ip.props.method = NM.SETTING_IP6_CONFIG_METHOD_AUTO
    elif is_dhcp and not is_autoconf:
        setting_ip.props.method = NM.SETTING_IP6_CONFIG_METHOD_DHCP


def _set_static(setting_ip, ip_addresses):
    for address in ip_addresses:
        if iplib.is_ipv6_link_local_addr(
            address[InterfaceIPv6.ADDRESS_IP],
            address[InterfaceIPv6.ADDRESS_PREFIX_LENGTH],
        ):
            logging.warning(
                "IPv6 link local address "
                "{a[ip]}/{a[prefix-length]} is ignored "
                "when applying desired state".format(a=address)
            )
        else:
            naddr = NM.IPAddress.new(
                socket.AF_INET6,
                address[InterfaceIPv6.ADDRESS_IP],
                address[InterfaceIPv6.ADDRESS_PREFIX_LENGTH],
            )
            setting_ip.add_address(naddr)

    if setting_ip.props.addresses:
        setting_ip.props.method = NM.SETTING_IP6_CONFIG_METHOD_MANUAL
    else:
        setting_ip.props.method = NM.SETTING_IP6_CONFIG_METHOD_LINK_LOCAL


def get_ip_profile(active_connection):
    """
    Get NMSettingIP6Config from NMActiveConnection.
    For any error, return None.
    """
    remote_conn = active_connection.get_connection()
    if remote_conn:
        return remote_conn.get_setting_ip6_config()
    return None


def acs_and_ip_profiles(nm_client):
    for ac in nm_client.get_active_connections():
        ip_profile = get_ip_profile(ac)
        if not ip_profile:
            continue
        yield ac, ip_profile


def is_dynamic(active_connection):
    ip_profile = get_ip_profile(active_connection)
    if ip_profile:
        method = ip_profile.get_method()
        return method in (
            NM.SETTING_IP6_CONFIG_METHOD_AUTO,
            NM.SETTING_IP6_CONFIG_METHOD_DHCP,
        )
    return False



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
import argparse
import errno
import fnmatch
import json
import logging
import os
import subprocess
import sys
import tempfile

import yaml

import libnmstate
from libnmstate import PrettyState
from libnmstate.error import NmstateConflictError
from libnmstate.error import NmstatePermissionError
from libnmstate.error import NmstateValueError
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIP
from libnmstate.schema import Route
from libnmstate.schema import RouteRule
from nmstatectl.nmstate_varlink import start_varlink_server


def main():
    logging.basicConfig(
        format="%(asctime)s %(name)-12s %(levelname)-8s %(message)s",
        level=logging.DEBUG,
    )

    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers()
    setup_subcommand_commit(subparsers)
    setup_subcommand_edit(subparsers)
    setup_subcommand_rollback(subparsers)
    setup_subcommand_set(subparsers)
    setup_subcommand_show(subparsers)
    setup_subcommand_version(subparsers)
    setup_subcommand_varlink(subparsers)
    parser.add_argument(
        "--version", action="store_true", help="Display nmstate version"
    )

    if len(sys.argv) == 1:
        parser.print_usage()
        return errno.EINVAL
    args = parser.parse_args()
    if args.version:
        print(libnmstate.__version__)
    else:
        return args.func(args)


def setup_subcommand_commit(subparsers):
    parser_commit = subparsers.add_parser("commit", help="Commit a change")
    parser_commit.add_argument(
        "checkpoint", nargs="?", default=None, help="checkpoint to commit"
    )
    parser_commit.set_defaults(func=commit)


def setup_subcommand_edit(subparsers):
    parser_edit = subparsers.add_parser(
        "edit", help="Edit network state in EDITOR"
    )
    parser_edit.set_defaults(func=edit)
    parser_edit.add_argument(
        "--json",
        help="Edit as JSON",
        default=True,
        action="store_false",
        dest="yaml",
    )
    parser_edit.add_argument(
        "only",
        default="*",
        nargs="?",
        metavar=Interface.KEY,
        help="Edit only specified interfaces (comma-separated)",
    )
    parser_edit.add_argument(
        "--no-verify",
        action="store_false",
        dest="verify",
        default=True,
        help="Do not verify that the state was completely set and disable "
        "rollback to previous state.",
    )
    parser_edit.add_argument(
        "--memory-only",
        action="store_false",
        dest="save_to_disk",
        default=True,
        help="Do not make the state persistent.",
    )


def setup_subcommand_rollback(subparsers):
    parser_rollback = subparsers.add_parser(
        "rollback", help="Rollback a change"
    )
    parser_rollback.add_argument(
        "checkpoint", nargs="?", default=None, help="checkpoint to roll back"
    )
    parser_rollback.set_defaults(func=rollback)


def setup_subcommand_set(subparsers):
    parser_set = subparsers.add_parser("set", help="Set network state")
    parser_set.add_argument(
        "file",
        help="File containing desired state. "
        "stdin is used when no file is specified.",
        nargs="*",
    )
    parser_set.add_argument(
        "--no-verify",
        action="store_false",
        dest="verify",
        default=True,
        help="Do not verify that the state was completely set and disable "
        "rollback to previous state",
    )
    parser_set.add_argument(
        "--no-commit",
        action="store_false",
        dest="commit",
        default=True,
        help="Do not commit new state after verification",
    )
    parser_set.add_argument(
        "--timeout",
        type=int,
        default=60,
        help="Timeout in seconds before reverting uncommited changes.",
    )
    parser_set.add_argument(
        "--memory-only",
        action="store_false",
        dest="save_to_disk",
        default=True,
        help="Do not make the state persistent.",
    )
    parser_set.set_defaults(func=apply)


def setup_subcommand_show(subparsers):
    parser_show = subparsers.add_parser("show", help="Show network state")
    parser_show.set_defaults(func=show)
    parser_show.add_argument(
        "--json",
        help="Edit as JSON",
        default=True,
        action="store_false",
        dest="yaml",
    )
    parser_show.add_argument(
        "-r, --running-config",
        help="Show running configurations",
        default=False,
        action="store_true",
        dest="running_config",
    )
    parser_show.add_argument(
        "-s, --saved-config",
        help="Show saved configurations",
        default=False,
        action="store_true",
        dest="saved_config",
    )
    parser_show.add_argument(
        "only",
        default="*",
        nargs="?",
        metavar=Interface.KEY,
        help="Show only specified interfaces (comma-separated)",
    )


def setup_subcommand_version(subparsers):
    parser_version = subparsers.add_parser(
        "version", help="Display nmstate version"
    )
    parser_version.set_defaults(func=version)


def setup_subcommand_varlink(subparsers):
    parser_varlink = subparsers.add_parser(
        "varlink", help="Varlink support for libnmstate"
    )
    parser_varlink.add_argument(
        "address", type=str, help="Unix socket address e.g: /run/nmstate"
    )
    parser_varlink.set_defaults(func=run_varlink_server)


def version(args):
    print(libnmstate.__version__)


def commit(args):
    try:
        libnmstate.commit(checkpoint=args.checkpoint)
    except NmstateValueError as e:
        print("ERROR committing change: {}\n".format(str(e)))
        return os.EX_DATAERR


def edit(args):
    state = _filter_state(libnmstate.show(), args.only)

    if not state[Interface.KEY]:
        sys.stderr.write("ERROR: No such interface\n")
        return os.EX_USAGE

    pretty_state = PrettyState(state)

    if args.yaml:
        suffix = ".yaml"
        txtstate = pretty_state.yaml
    else:
        suffix = ".json"
        txtstate = pretty_state.json

    new_state = _get_edited_state(txtstate, suffix, args.yaml)
    if not new_state:
        return os.EX_DATAERR

    print("Applying the following state: ")
    print_state(new_state, use_yaml=args.yaml)

    libnmstate.apply(
        new_state, verify_change=args.verify, save_to_disk=args.save_to_disk
    )


def rollback(args):
    try:
        libnmstate.rollback(checkpoint=args.checkpoint)
    except NmstateValueError as e:
        print("ERROR rolling back change: {}\n".format(str(e)))
        return os.EX_DATAERR


def show(args):
    if args.running_config:
        full_state = libnmstate.show_running_config()
    elif args.saved_config:
        full_state = libnmstate.show_saved_config()
    else:
        full_state = libnmstate.show()
    state = _filter_state(full_state, args.only)
    print_state(state, use_yaml=args.yaml)


def apply(args):
    if args.file:
        for statefile in args.file:
            if statefile == "-" and not os.path.isfile(statefile):
                statedata = sys.stdin.read()
            else:
                with open(statefile) as statefile:
                    statedata = statefile.read()

            ret = apply_state(
                statedata,
                args.verify,
                args.commit,
                args.timeout,
                args.save_to_disk,
            )
            if ret:
                return ret
    elif not sys.stdin.isatty():
        statedata = sys.stdin.read()
        return apply_state(
            statedata,
            args.verify,
            args.commit,
            args.timeout,
            args.save_to_disk,
        )
    else:
        sys.stderr.write("ERROR: No state specified\n")
        return 1


def run_varlink_server(args):
    try:
        start_varlink_server(args.address)
    except Exception as exception:
        logging.exception(exception)


def apply_state(statedata, verify_change, commit, timeout, save_to_disk):
    use_yaml = False
    # JSON dictionaries start with a curly brace
    if statedata[0] == "{":
        state = json.loads(statedata)
    else:
        state = yaml.load(statedata, Loader=yaml.SafeLoader)
        use_yaml = True

    try:
        checkpoint = libnmstate.apply(
            state,
            verify_change=verify_change,
            commit=commit,
            rollback_timeout=timeout,
            save_to_disk=save_to_disk,
        )
    except NmstatePermissionError as e:
        sys.stderr.write("ERROR: Missing permissions:{}\n".format(str(e)))
        return os.EX_NOPERM
    except NmstateConflictError:
        sys.stderr.write(
            "ERROR: State editing already in progress.\n"
            "Commit, roll back or wait before retrying.\n"
        )
        return os.EX_UNAVAILABLE

    print("Desired state applied: ")
    print_state(state, use_yaml=use_yaml)
    if checkpoint:
        print("Checkpoint: {}".format(checkpoint))


def _filter_state(state, whitelist):
    if whitelist != "*":
        patterns = [p for p in whitelist.split(",")]
        state[Interface.KEY] = _filter_interfaces(state, patterns)
        state[Route.KEY] = _filter_routes(state, patterns)
        state[RouteRule.KEY] = _filter_route_rule(state, patterns)
    return state


def _filter_interfaces(state, patterns):
    """
    return the states for all interfaces from `state` that match at least one
    of the provided patterns.
    """
    showinterfaces = []

    for interface in state[Interface.KEY]:
        for pattern in patterns:
            if fnmatch.fnmatch(interface["name"], pattern):
                showinterfaces.append(interface)
                break
    return showinterfaces


def _get_edited_state(txtstate, suffix, use_yaml):
    while True:
        txtstate = _run_editor(txtstate, suffix)

        if txtstate is None:
            return None

        new_state, error = _parse_state(txtstate, use_yaml)

        if error:
            if not _try_edit_again(error):
                return None
        else:
            return new_state


def _run_editor(txtstate, suffix):
    editor = os.environ.get("EDITOR", "vi")
    with tempfile.NamedTemporaryFile(
        suffix=suffix, prefix="nmstate-"
    ) as statefile:
        statefile.write(txtstate.encode("utf-8"))
        statefile.flush()

        try:
            subprocess.check_call([editor, statefile.name])
            statefile.seek(0)
            return statefile.read()

        except subprocess.CalledProcessError:
            sys.stderr.write("Error running editor, aborting...\n")
            return None


def _parse_state(txtstate, parse_yaml):
    error = ""
    state = {}
    if parse_yaml:
        try:
            state = yaml.load(txtstate, Loader=yaml.SafeLoader)
        except yaml.parser.ParserError as e:
            error = "Invalid YAML syntax: %s\n" % e
        except yaml.parser.ScannerError as e:
            error = "Invalid YAML syntax: %s\n" % e
    else:
        try:
            state = json.loads(txtstate)
        except ValueError as e:
            error = "Invalid JSON syntax: %s\n" % e

    if not error and Interface.KEY not in state:
        # Allow editing routes only.
        state[Interface.KEY] = []

    return state, error


def _try_edit_again(error):
    """
    Print error and ask for user feedback. Return True, if the state should be
    edited again and False otherwise.
    """

    sys.stderr.write("ERROR: " + error)
    response = ""
    while response not in ("y", "n"):
        response = input(
            "Try again? [y,n]:\n"
            "y - yes, start editor again\n"
            "n - no, throw away my changes\n"
            "> "
        ).lower()
        if response == "n":
            return False
    return True


def print_state(state, use_yaml=False):
    state = PrettyState(state)
    if use_yaml:
        sys.stdout.write(state.yaml)
    else:
        print(state.json)


def _filter_routes(state, patterns):
    """
    return the states for all routes from `state` that match at least one
    of the provided patterns.
    """
    routes = {}
    for route_type in state.get(Route.KEY, {}).keys():
        routes[route_type] = []
        for route in state.get(Route.KEY, {}).get(route_type, []):
            for pattern in patterns:
                if fnmatch.fnmatch(route[Route.NEXT_HOP_INTERFACE], pattern):
                    routes[route_type].append(route)
    return routes


def _filter_route_rule(state, patterns):
    route_rules = _filter_route_rule_by_table_id(state)
    _filter_route_rule_by_autorule_table_id(state, patterns, route_rules)
    return route_rules


def _filter_route_rule_by_table_id(state):
    """
    return the rules for state's route rule that match the table_id of the
    filtered route of state by interface
    """
    route_rules = {RouteRule.CONFIG: []}
    table_ids = []
    for routes in {Route.CONFIG: [], Route.RUNNING: []}:
        for route in state.get(Route.KEY, {}).get(routes, []):
            if route.get(Route.TABLE_ID) not in table_ids:
                table_ids.append(route.get(Route.TABLE_ID))
    for rule in state.get(RouteRule.KEY, {}).get(RouteRule.CONFIG, []):
        if rule.get(RouteRule.ROUTE_TABLE) in table_ids:
            route_rules[RouteRule.CONFIG].append(rule)
    return route_rules


def _filter_route_rule_by_autorule_table_id(state, patterns, route_rules):
    """
    return the rules that match the iface's name and AUTO_ROUTE_TABLE_ID
    """
    table_ids = []
    for pattern in patterns:
        for interface in state[Interface.KEY]:
            if fnmatch.fnmatch(interface[Interface.NAME], pattern):
                autotable = interface.get(Interface.IPV4, {}).get(
                    InterfaceIP.AUTO_ROUTE_TABLE_ID
                )
                if autotable:
                    table_ids.append(autotable)
                autotable = interface.get(Interface.IPV6, {}).get(
                    InterfaceIP.AUTO_ROUTE_TABLE_ID
                )
                if autotable:
                    table_ids.append(autotable)
    for rule in state.get(RouteRule.KEY, {}).get(RouteRule.CONFIG, []):
        if (
            rule.get(RouteRule.ROUTE_TABLE) in table_ids
            and rule not in route_rules[RouteRule.CONFIG]
        ):
            route_rules[RouteRule.CONFIG].append(rule)



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import pytest

from libnmstate.ifaces.linux_bridge_port_vlan import NmstateLinuxBridgePortVlan
from libnmstate.schema import LinuxBridge as LB


@pytest.mark.parametrize(
    "is_native_vlan", [True, False], ids=["native", "not-native"]
)
def test_bridge_vlan_trunk_port(is_native_vlan):
    info = {
        LB.Port.Vlan.MODE: LB.Port.Vlan.Mode.TRUNK,
        LB.Port.Vlan.TRUNK_TAGS: [
            {LB.Port.Vlan.TrunkTags.ID: 100},
            {
                LB.Port.Vlan.TrunkTags.ID_RANGE: {
                    LB.Port.Vlan.TrunkTags.MIN_RANGE: 200,
                    LB.Port.Vlan.TrunkTags.MAX_RANGE: 300,
                }
            },
            {
                LB.Port.Vlan.TrunkTags.ID_RANGE: {
                    LB.Port.Vlan.TrunkTags.MIN_RANGE: 300,
                    LB.Port.Vlan.TrunkTags.MAX_RANGE: 400,
                }
            },
        ],
        LB.Port.Vlan.ENABLE_NATIVE: is_native_vlan,
    }
    if is_native_vlan:
        info[LB.Port.Vlan.TAG] = 10
    nmstate_port = NmstateLinuxBridgePortVlan(info)

    kernel_vlans = nmstate_port.to_kernel_vlans()
    new_nmstate_port = NmstateLinuxBridgePortVlan.new_from_kernel_vlans(
        kernel_vlans
    )

    assert nmstate_port.to_dict() == new_nmstate_port.to_dict()


def test_bridge_vlan_access_port():
    info = {LB.Port.Vlan.MODE: LB.Port.Vlan.Mode.ACCESS, LB.Port.Vlan.TAG: 100}
    nmstate_port = NmstateLinuxBridgePortVlan(info)

    kernel_vlans = nmstate_port.to_kernel_vlans()
    new_nmstate_port = NmstateLinuxBridgePortVlan.new_from_kernel_vlans(
        kernel_vlans
    )

    assert nmstate_port.to_dict() == new_nmstate_port.to_dict()


def test_bridge_vlan_trunk_port_range_to_flat():
    info = {
        LB.Port.Vlan.ENABLE_NATIVE: False,
        LB.Port.Vlan.MODE: LB.Port.Vlan.Mode.TRUNK,
        LB.Port.Vlan.TRUNK_TAGS: [
            {
                LB.Port.Vlan.TrunkTags.ID_RANGE: {
                    LB.Port.Vlan.TrunkTags.MIN_RANGE: 101,
                    LB.Port.Vlan.TrunkTags.MAX_RANGE: 102,
                }
            },
            {
                LB.Port.Vlan.TrunkTags.ID_RANGE: {
                    LB.Port.Vlan.TrunkTags.MIN_RANGE: 103,
                    LB.Port.Vlan.TrunkTags.MAX_RANGE: 103,
                }
            },
        ],
    }
    nmstate_port = NmstateLinuxBridgePortVlan(info)
    expected_info = {
        LB.Port.Vlan.ENABLE_NATIVE: False,
        LB.Port.Vlan.MODE: LB.Port.Vlan.Mode.TRUNK,
        LB.Port.Vlan.TRUNK_TAGS: [
            {LB.Port.Vlan.TrunkTags.ID: 101},
            {LB.Port.Vlan.TrunkTags.ID: 102},
            {LB.Port.Vlan.TrunkTags.ID: 103},
        ],
    }
    assert nmstate_port.to_dict(expand_vlan_range=True) == expected_info



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from copy import deepcopy

import pytest

from libnmstate.error import NmstateVerificationError
from libnmstate.error import NmstateNotImplementedError
from libnmstate.schema import DNS
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceIPv6
from libnmstate.schema import Route

from libnmstate.ifaces import BaseIface
from libnmstate.ifaces import Ifaces
from libnmstate.route import RouteState
from libnmstate.dns import DnsState

from .testlib.ifacelib import gen_foo_iface_info
from .testlib.ifacelib import gen_two_static_ip_ifaces
from .testlib.routelib import IPV4_ROUTE_IFACE_NAME
from .testlib.routelib import IPV6_ROUTE_IFACE_NAME
from .testlib.routelib import gen_ipv4_default_gateway
from .testlib.routelib import gen_ipv6_default_gateway


IPV4_DNS_SERVER1 = "192.0.2.1"
IPV4_DNS_SERVER2 = "192.0.2.2"
IPV4_DNS_SERVER3 = "192.0.2.3"
IPV6_DNS_SERVER1 = "2001:db8:a::1"
IPV6_DNS_SERVER2 = "2001:db8:a::2"
IPV6_DNS_SERVER3 = "2001:db8:a::3"

DNS_SEARCHES_1 = ["example.com", "example.org"]
DNS_SEARCHES_2 = ["example.info", "example.net"]

DNS_CONFIG1 = {
    DNS.SERVER: [IPV4_DNS_SERVER1, IPV6_DNS_SERVER1],
    DNS.SEARCH: DNS_SEARCHES_1,
}

DNS_CONFIG2 = {
    DNS.SERVER: [IPV6_DNS_SERVER2, IPV4_DNS_SERVER2],
    DNS.SEARCH: DNS_SEARCHES_2,
}


class TestDnsState:
    def test_merge_without_desire_with_current(self):
        dns_state = DnsState({}, {DNS.CONFIG: DNS_CONFIG1})

        assert dns_state.config == DNS_CONFIG1
        assert dns_state.current_config == DNS_CONFIG1

    def test_merge_by_overriding_current(self):
        dns_state = DnsState(
            {DNS.CONFIG: DNS_CONFIG2},
            {DNS.CONFIG: DNS_CONFIG1},
        )

        assert dns_state.config == DNS_CONFIG2
        assert dns_state.current_config == DNS_CONFIG1

    def test_merge_by_discarding_current(self):
        dns_state = DnsState({DNS.CONFIG: []}, {DNS.CONFIG: DNS_CONFIG1})

        assert dns_state.config == {DNS.SERVER: [], DNS.SEARCH: []}
        assert dns_state.current_config == DNS_CONFIG1

    def test_gen_metadadata_use_default_gateway_ipv4_server_prefered(self):
        ifaces = self._gen_static_ifaces()
        route_state = self._gen_route_state(ifaces)
        dns_state = DnsState({DNS.CONFIG: DNS_CONFIG1}, {})
        ifaces.gen_dns_metadata(dns_state, route_state)
        ipv4_iface = ifaces.all_kernel_ifaces[IPV4_ROUTE_IFACE_NAME]
        ipv6_iface = ifaces.all_kernel_ifaces[IPV6_ROUTE_IFACE_NAME]

        assert dns_state.config == DNS_CONFIG1
        assert ipv4_iface.to_dict()[Interface.IPV4][
            BaseIface.DNS_METADATA
        ] == {
            DnsState.PRIORITY_METADATA: 0,
            DNS.SERVER: [IPV4_DNS_SERVER1],
            DNS.SEARCH: DNS_SEARCHES_1,
        }
        assert ipv6_iface.to_dict()[Interface.IPV6][
            BaseIface.DNS_METADATA
        ] == {
            DnsState.PRIORITY_METADATA: 1,
            DNS.SERVER: [IPV6_DNS_SERVER1],
            DNS.SEARCH: [],
        }

    def test_gen_metadadata_use_default_gateway_ipv6_server_prefered(self):
        ifaces = self._gen_static_ifaces()
        route_state = self._gen_route_state(ifaces)
        dns_state = DnsState({DNS.CONFIG: DNS_CONFIG2}, {})
        ifaces.gen_dns_metadata(dns_state, route_state)
        ipv4_iface = ifaces.all_kernel_ifaces[IPV4_ROUTE_IFACE_NAME]
        ipv6_iface = ifaces.all_kernel_ifaces[IPV6_ROUTE_IFACE_NAME]

        assert dns_state.config == DNS_CONFIG2
        assert ipv4_iface.to_dict()[Interface.IPV4][
            BaseIface.DNS_METADATA
        ] == {
            DnsState.PRIORITY_METADATA: 1,
            DNS.SERVER: [IPV4_DNS_SERVER2],
            DNS.SEARCH: [],
        }
        assert ipv6_iface.to_dict()[Interface.IPV6][
            BaseIface.DNS_METADATA
        ] == {
            DnsState.PRIORITY_METADATA: 0,
            DNS.SERVER: [IPV6_DNS_SERVER2],
            DNS.SEARCH: DNS_SEARCHES_2,
        }

    def test_gen_metadata_use_dynamic_interface(self):
        ifaces = self._gen_dynamic_ifaces()
        route_state = self._gen_route_state(ifaces)
        dns_state = DnsState({DNS.CONFIG: DNS_CONFIG1}, {})
        ifaces.gen_dns_metadata(dns_state, route_state)
        ipv4_iface = ifaces.all_kernel_ifaces[IPV4_ROUTE_IFACE_NAME]
        ipv6_iface = ifaces.all_kernel_ifaces[IPV6_ROUTE_IFACE_NAME]

        assert dns_state.config == DNS_CONFIG1
        assert ipv4_iface.to_dict()[Interface.IPV4][
            BaseIface.DNS_METADATA
        ] == {
            DnsState.PRIORITY_METADATA: 0,
            DNS.SERVER: [IPV4_DNS_SERVER1],
            DNS.SEARCH: DNS_SEARCHES_1,
        }
        assert ipv6_iface.to_dict()[Interface.IPV6][
            BaseIface.DNS_METADATA
        ] == {
            DnsState.PRIORITY_METADATA: 1,
            DNS.SERVER: [IPV6_DNS_SERVER1],
            DNS.SEARCH: [],
        }

    @pytest.fixture
    def with_dns_metadata(self):
        ifaces = self._gen_static_ifaces()
        route_state = self._gen_route_state(ifaces)
        dns_state = DnsState({}, {DNS.CONFIG: DNS_CONFIG1})
        ifaces.gen_dns_metadata(dns_state, route_state)
        ipv4_iface = ifaces.all_kernel_ifaces[IPV4_ROUTE_IFACE_NAME]
        ipv6_iface = ifaces.all_kernel_ifaces[IPV6_ROUTE_IFACE_NAME]

        assert dns_state.config == DNS_CONFIG1
        assert ipv4_iface.to_dict()[Interface.IPV4][
            BaseIface.DNS_METADATA
        ] == {
            DnsState.PRIORITY_METADATA: 0,
            DNS.SERVER: [IPV4_DNS_SERVER1],
            DNS.SEARCH: DNS_SEARCHES_1,
        }
        assert ipv6_iface.to_dict()[Interface.IPV6][
            BaseIface.DNS_METADATA
        ] == {
            DnsState.PRIORITY_METADATA: 1,
            DNS.SERVER: [IPV6_DNS_SERVER1],
            DNS.SEARCH: [],
        }
        yield ipv4_iface, ipv6_iface

    def test_remove_metadata(self, with_dns_metadata):
        ipv4_iface, ipv6_iface = with_dns_metadata
        ipv4_iface.remove_dns_metadata()
        ipv6_iface.remove_dns_metadata()

        assert (
            BaseIface.DNS_METADATA not in ipv4_iface.to_dict()[Interface.IPV4]
        )
        assert (
            BaseIface.DNS_METADATA not in ipv6_iface.to_dict()[Interface.IPV6]
        )

    def test_verify_identical_state(self):
        dns_state = DnsState({DNS.CONFIG: DNS_CONFIG1}, {})
        dns_state.verify({DNS.CONFIG: DNS_CONFIG1})

    def test_verify_not_match(self):
        dns_state = DnsState({DNS.CONFIG: DNS_CONFIG1}, {})
        with pytest.raises(NmstateVerificationError):
            dns_state.verify({DNS.CONFIG: DNS_CONFIG2})

    @pytest.mark.parametrize(
        "dns_servers",
        [
            ([IPV4_DNS_SERVER1, IPV4_DNS_SERVER2, IPV4_DNS_SERVER3]),
            ([IPV6_DNS_SERVER1, IPV6_DNS_SERVER2, IPV6_DNS_SERVER3]),
            ([IPV4_DNS_SERVER1, IPV4_DNS_SERVER2, IPV6_DNS_SERVER1]),
            ([IPV6_DNS_SERVER1, IPV6_DNS_SERVER2, IPV4_DNS_SERVER1]),
            pytest.param(
                ([IPV4_DNS_SERVER1, IPV6_DNS_SERVER1, IPV4_DNS_SERVER2]),
                marks=pytest.mark.xfail(
                    reason="Not supported",
                    raises=NmstateNotImplementedError,
                    strict=True,
                ),
            ),
            pytest.param(
                ([IPV6_DNS_SERVER1, IPV4_DNS_SERVER1, IPV6_DNS_SERVER2]),
                marks=pytest.mark.xfail(
                    reason="Not supported",
                    raises=NmstateNotImplementedError,
                    strict=True,
                ),
            ),
            (
                [
                    IPV4_DNS_SERVER1,
                    IPV4_DNS_SERVER2,
                    IPV6_DNS_SERVER1,
                    IPV6_DNS_SERVER2,
                ]
            ),
            (
                [
                    IPV6_DNS_SERVER1,
                    IPV6_DNS_SERVER2,
                    IPV4_DNS_SERVER1,
                    IPV4_DNS_SERVER2,
                ]
            ),
        ],
        ids=[
            "3ipv4",
            "3ipv6",
            "2ipv4+ipv6",
            "2ipv6+ipv4",
            "ipv4+ipv6+ipv4",
            "ipv6+ipv4+ipv6",
            "2ipv4+2ipv6",
            "2ipv6+2ipv4",
        ],
    )
    def test_validate_3_more_name_servers(self, dns_servers):
        dns_config = deepcopy(DNS_CONFIG1)
        dns_config[DNS.SERVER] = dns_servers
        DnsState({DNS.CONFIG: dns_config}, {})

    def test_3_dns_ipv4_servers(self):
        dns_config = deepcopy(DNS_CONFIG1)
        dns_config[DNS.SERVER] = [
            IPV4_DNS_SERVER1,
            IPV4_DNS_SERVER2,
            IPV4_DNS_SERVER3,
        ]

        dns_state = DnsState({DNS.CONFIG: dns_config}, {})
        ifaces = self._gen_static_ifaces()
        route_state = self._gen_route_state(ifaces)
        ifaces.gen_dns_metadata(dns_state, route_state)

        ipv4_iface = ifaces.all_kernel_ifaces[IPV4_ROUTE_IFACE_NAME]

        assert dns_state.config == dns_config
        assert ipv4_iface.to_dict()[Interface.IPV4][
            BaseIface.DNS_METADATA
        ] == {
            DnsState.PRIORITY_METADATA: 0,
            DNS.SERVER: dns_config[DNS.SERVER],
            DNS.SEARCH: DNS_SEARCHES_1,
        }

    def test_3_dns_ipv6_servers(self):
        dns_config = deepcopy(DNS_CONFIG1)
        dns_config[DNS.SERVER] = [
            IPV6_DNS_SERVER1,
            IPV6_DNS_SERVER2,
            IPV6_DNS_SERVER3,
        ]

        ifaces = self._gen_static_ifaces()
        route_state = self._gen_route_state(ifaces)
        dns_state = DnsState({DNS.CONFIG: dns_config}, {})
        ifaces.gen_dns_metadata(dns_state, route_state)
        ipv6_iface = ifaces.all_kernel_ifaces[IPV6_ROUTE_IFACE_NAME]

        assert dns_state.config == dns_config
        assert ipv6_iface.to_dict()[Interface.IPV6][
            BaseIface.DNS_METADATA
        ] == {
            DnsState.PRIORITY_METADATA: 0,
            DNS.SERVER: dns_config[DNS.SERVER],
            DNS.SEARCH: DNS_SEARCHES_1,
        }

    def _gen_dynamic_ifaces(self):
        ipv4_iface_info = gen_foo_iface_info()
        ipv4_iface_info[Interface.NAME] = IPV4_ROUTE_IFACE_NAME
        ipv4_iface_info[Interface.IPV4] = {
            InterfaceIPv4.ENABLED: True,
            InterfaceIPv4.DHCP: True,
            InterfaceIPv4.AUTO_DNS: False,
        }
        ipv6_iface_info = gen_foo_iface_info()
        ipv6_iface_info[Interface.NAME] = IPV6_ROUTE_IFACE_NAME
        ipv6_iface_info[Interface.IPV6] = {
            InterfaceIPv6.ENABLED: True,
            InterfaceIPv6.DHCP: True,
            InterfaceIPv6.AUTOCONF: True,
            InterfaceIPv6.AUTO_DNS: False,
        }
        return Ifaces([], [ipv4_iface_info, ipv6_iface_info])

    def _gen_static_ifaces(self):
        return gen_two_static_ip_ifaces(
            IPV4_ROUTE_IFACE_NAME, IPV6_ROUTE_IFACE_NAME
        )

    def _gen_route_state(self, ifaces):
        return RouteState(
            ifaces,
            {},
            {
                Route.CONFIG: [
                    gen_ipv4_default_gateway().to_dict(),
                    gen_ipv6_default_gateway().to_dict(),
                ]
            },
        )



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
import copy

import pytest

from unittest import mock

from libnmstate import netapplier
from libnmstate.schema import Bond
from libnmstate.schema import BondMode
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceIPv6
from libnmstate.schema import InterfaceState
from libnmstate.schema import InterfaceType

BOND_TYPE = InterfaceType.BOND


@pytest.fixture
def show_with_plugins_mock():
    with mock.patch.object(netapplier, "show_with_plugins") as m:
        yield m


@pytest.fixture
def plugin_context_mock():
    with mock.patch.object(netapplier, "plugin_context") as m:
        #        def enter(self):
        #            return [mock.MagicMock()]
        #
        #        m().__enter__ = enter
        yield m


@pytest.fixture
def net_state_mock():
    with mock.patch.object(netapplier, "NetState") as m:
        yield m


def test_iface_admin_state_change(
    show_with_plugins_mock,
    plugin_context_mock,
    net_state_mock,
):
    current_config = {
        Interface.KEY: [
            {
                Interface.NAME: "foo",
                Interface.TYPE: InterfaceType.DUMMY,
                Interface.STATE: InterfaceState.UP,
                Interface.IPV4: {InterfaceIPv4.ENABLED: False},
                Interface.IPV6: {InterfaceIPv6.ENABLED: False},
            }
        ]
    }
    desired_config = copy.deepcopy(current_config)

    desired_config[Interface.KEY][0][Interface.STATE] = InterfaceState.DOWN
    show_with_plugins_mock.return_value = current_config
    plugin = mock.MagicMock()
    plugin_context_mock.return_value.__enter__.return_value = [plugin]
    netapplier.apply(desired_config, verify_change=False)

    plugin.apply_changes.assert_called_once_with(
        net_state_mock(desired_config, current_config), True
    )


def test_add_new_bond(
    plugin_context_mock,
    show_with_plugins_mock,
    net_state_mock,
):
    show_with_plugins_mock.return_value = {}

    desired_config = {
        Interface.KEY: [
            {
                Interface.NAME: "bond99",
                Interface.TYPE: BOND_TYPE,
                Interface.STATE: InterfaceState.UP,
                Bond.CONFIG_SUBTREE: {
                    Bond.MODE: BondMode.ROUND_ROBIN,
                    Bond.PORT: [],
                    Bond.OPTIONS_SUBTREE: {"miimon": 200},
                },
                Interface.IPV4: {},
                Interface.IPV6: {},
            }
        ]
    }

    plugin = mock.MagicMock()
    plugin_context_mock.return_value.__enter__.return_value = [plugin]
    netapplier.apply(desired_config, verify_change=False)

    plugin.apply_changes.assert_called_once_with(
        net_state_mock(desired_config, {}), True
    )


def test_edit_existing_bond(
    show_with_plugins_mock,
    plugin_context_mock,
    net_state_mock,
):
    current_config = {
        Interface.KEY: [
            {
                Interface.NAME: "bond99",
                Interface.TYPE: BOND_TYPE,
                Interface.STATE: InterfaceState.UP,
                Bond.CONFIG_SUBTREE: {
                    Bond.MODE: BondMode.ROUND_ROBIN,
                    Bond.PORT: [],
                    Bond.OPTIONS_SUBTREE: {"miimon": "100"},
                },
                Interface.IPV4: {InterfaceIPv4.ENABLED: False},
                Interface.IPV6: {InterfaceIPv6.ENABLED: False},
            }
        ]
    }
    show_with_plugins_mock.return_value = current_config

    desired_config = copy.deepcopy(current_config)
    options = desired_config[Interface.KEY][0][Bond.CONFIG_SUBTREE][
        Bond.OPTIONS_SUBTREE
    ]
    options["miimon"] = 200

    plugin = mock.MagicMock()
    plugin_context_mock.return_value.__enter__.return_value = [plugin]
    netapplier.apply(desired_config, verify_change=False)

    plugin.apply_changes.assert_called_once_with(
        net_state_mock(desired_config, current_config), True
    )


def test_error_apply():
    with pytest.raises(TypeError):
        # pylint: disable=too-many-function-args
        netapplier.apply({"interfaces": []}, True)
        # pylint: enable=too-many-function-args

    with pytest.raises(TypeError):
        # pylint: disable=too-many-function-args
        netapplier.apply({"interfaces": []}, True, True, 0)
        # pylint: enable=too-many-function-args


def test_error_commit():
    with pytest.raises(TypeError):
        # pylint: disable=too-many-function-args
        netapplier.commit(None)
        # pylint: enable=too-many-function-args


def test_error_rollback():
    with pytest.raises(TypeError):
        # pylint: disable=too-many-function-args
        netapplier.rollback(None)
        # pylint: enable=too-many-function-args
